#!/usr/bin/env python3

# Allow direct execution
import os
import sys
import unittest

sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.abspath(__file__))))


import subprocess

from yt_dlp import YoutubeDL
from yt_dlp.utils import shell_quote
from yt_dlp.postprocessor import (
    ExecPP,
    FFmpegThumbnailsConvertorPP,
    MetadataFromFieldPP,
    MetadataParserPP,
    ModifyChaptersPP,
    SponsorBlockPP,
)


class TestMetadataFromField(unittest.TestCase):

    def test_format_to_regex(self):
        self.assertEqual(
            MetadataParserPP.format_to_regex('%(title)s - %(artist)s'),
            r'(?P<title>.+)\ \-\ (?P<artist>.+)')
        self.assertEqual(MetadataParserPP.format_to_regex(r'(?P<x>.+)'), r'(?P<x>.+)')

    def test_field_to_template(self):
        self.assertEqual(MetadataParserPP.field_to_template('title'), '%(title)s')
        self.assertEqual(MetadataParserPP.field_to_template('1'), '1')
        self.assertEqual(MetadataParserPP.field_to_template('foo bar'), 'foo bar')
        self.assertEqual(MetadataParserPP.field_to_template(' literal'), ' literal')

    def test_metadatafromfield(self):
        self.assertEqual(
            MetadataFromFieldPP.to_action('%(title)s \\: %(artist)s:%(title)s : %(artist)s'),
            (MetadataParserPP.Actions.INTERPRET, '%(title)s : %(artist)s', '%(title)s : %(artist)s'))


class TestConvertThumbnail(unittest.TestCase):
    def test_escaping(self):
        pp = FFmpegThumbnailsConvertorPP()
        if not pp.available:
            print('Skipping: ffmpeg not found')
            return

        test_data_dir = 'test/testdata/thumbnails'
        generated_file = f'{test_data_dir}/empty.webp'

        subprocess.check_call([
            pp.executable, '-y', '-f', 'lavfi', '-i', 'color=c=black:s=320x320',
            '-c:v', 'libwebp', '-pix_fmt', 'yuv420p', '-vframes', '1', generated_file,
        ], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)

        file = test_data_dir + '/foo %d bar/foo_%d.{}'
        initial_file = file.format('webp')
        os.replace(generated_file, initial_file)

        tests = (('webp', 'png'), ('png', 'jpg'))

        for inp, out in tests:
            out_file = file.format(out)
            if os.path.exists(out_file):
                os.remove(out_file)
            pp.convert_thumbnail(file.format(inp), out)
            self.assertTrue(os.path.exists(out_file))

        for _, out in tests:
            os.remove(file.format(out))

        os.remove(initial_file)


class TestExec(unittest.TestCase):
    def test_parse_cmd(self):
        pp = ExecPP(YoutubeDL(), '')
        info = {'filepath': 'file name'}
        cmd = 'echo {}'.format(shell_quote(info['filepath']))

        self.assertEqual(pp.parse_cmd('echo', info), cmd)
        self.assertEqual(pp.parse_cmd('echo {}', info), cmd)
        self.assertEqual(pp.parse_cmd('echo %(filepath)q', info), cmd)


class TestModifyChaptersPP(unittest.TestCase):
    def setUp(self):
        self._pp = ModifyChaptersPP(YoutubeDL())

    @staticmethod
    def _sponsor_chapter(start, end, cat, remove=False, title=None):
        if title is None:
            title = SponsorBlockPP.CATEGORIES[cat]
        return {
            'start_time': start,
            'end_time': end,
            '_categories': [(cat, start, end, title)],
            **({'remove': True} if remove else {}),
        }

    @staticmethod
    def _chapter(start, end, title=None, remove=False):
        c = {'start_time': start, 'end_time': end}
        if title is not None:
            c['title'] = title
        if remove:
            c['remove'] = True
        return c

    def _chapters(self, ends, titles):
        self.assertEqual(len(ends), len(titles))
        start = 0
        chapters = []
        for e, t in zip(ends, titles):
            chapters.append(self._chapter(start, e, t))
            start = e
        return chapters

    def _remove_marked_arrange_sponsors_test_impl(
            self, chapters, expected_chapters, expected_removed):
        actual_chapters, actual_removed = (
            self._pp._remove_marked_arrange_sponsors(chapters))
        for c in actual_removed:
            c.pop('title', None)
            c.pop('_categories', None)
        actual_chapters = [{
            'start_time': c['start_time'],
            'end_time': c['end_time'],
            'title': c['title'],
        } for c in actual_chapters]
        self.assertSequenceEqual(expected_chapters, actual_chapters)
        self.assertSequenceEqual(expected_removed, actual_removed)

    def test_remove_marked_arrange_sponsors_CanGetThroughUnaltered(self):
        chapters = self._chapters([10, 20, 30, 40], ['c1', 'c2', 'c3', 'c4'])
        self._remove_marked_arrange_sponsors_test_impl(chapters, chapters, [])

    def test_remove_marked_arrange_sponsors_ChapterWithSponsors(self):
        chapters = [
            *self._chapters([70], ['c']),
            self._sponsor_chapter(10, 20, 'sponsor'),
            self._sponsor_chapter(30, 40, 'preview'),
            self._sponsor_chapter(50, 60, 'filler')]
        expected = self._chapters(
            [10, 20, 30, 40, 50, 60, 70],
            ['c', '[SponsorBlock]: Sponsor', 'c', '[SponsorBlock]: Preview/Recap',
             'c', '[SponsorBlock]: Filler Tangent', 'c'])
        self._remove_marked_arrange_sponsors_test_impl(chapters, expected, [])

    def test_remove_marked_arrange_sponsors_SponsorBlockChapters(self):
        chapters = [
            *self._chapters([70], ['c']),
            self._sponsor_chapter(10, 20, 'chapter', title='sb c1'),
            self._sponsor_chapter(15, 16, 'chapter', title='sb c2'),
            self._sponsor_chapter(30, 40, 'preview'),
            self._sponsor_chapter(50, 60, 'filler')]
        expected = self._chapters(
            [10, 15, 16, 20, 30, 40, 50, 60, 70],
            ['c', '[SponsorBlock]: sb c1', '[SponsorBlock]: sb c1, sb c2', '[SponsorBlock]: sb c1',
             'c', '[SponsorBlock]: Preview/Recap',
             'c', '[SponsorBlock]: Filler Tangent', 'c'])
        self._remove_marked_arrange_sponsors_test_impl(chapters, expected, [])

    def test_remove_marked_arrange_sponsors_UniqueNamesForOverlappingSponsors(self):
        chapters = [
            *self._chapters([120], ['c']),
            self._sponsor_chapter(10, 45, 'sponsor'),
            self._sponsor_chapter(20, 40, 'selfpromo'),
            self._sponsor_chapter(50, 70, 'sponsor'),
            self._sponsor_chapter(60, 85, 'selfpromo'),
            self._sponsor_chapter(90, 120, 'selfpromo'),
            self._sponsor_chapter(100, 110, 'sponsor')]
        expected = self._chapters(
            [10, 20, 40, 45, 50, 60, 70, 85, 90, 100, 110, 120],
            ['c', '[SponsorBlock]: Sponsor', '[SponsorBlock]: Sponsor, Unpaid/Self Promotion',
             '[SponsorBlock]: Sponsor',
             'c', '[SponsorBlock]: Sponsor', '[SponsorBlock]: Sponsor, Unpaid/Self Promotion',
             '[SponsorBlock]: Unpaid/Self Promotion',
             'c', '[SponsorBlock]: Unpaid/Self Promotion', '[SponsorBlock]: Unpaid/Self Promotion, Sponsor',
             '[SponsorBlock]: Unpaid/Self Promotion'])
        self._remove_marked_arrange_sponsors_test_impl(chapters, expected, [])

    def test_remove_marked_arrange_sponsors_ChapterWithCuts(self):
        cuts = [self._chapter(10, 20, remove=True),
                self._sponsor_chapter(30, 40, 'sponsor', remove=True),
                self._chapter(50, 60, remove=True)]
        chapters = self._chapters([70], ['c']) + cuts
        self._remove_marked_arrange_sponsors_test_impl(
            chapters, self._chapters([40], ['c']), cuts)

    def test_remove_marked_arrange_sponsors_ChapterWithSponsorsAndCuts(self):
        chapters = [
            *self._chapters([70], ['c']),
            self._sponsor_chapter(10, 20, 'sponsor'),
            self._sponsor_chapter(30, 40, 'selfpromo', remove=True),
            self._sponsor_chapter(50, 60, 'interaction')]
        expected = self._chapters([10, 20, 40, 50, 60],
                                  ['c', '[SponsorBlock]: Sponsor', 'c',
                                   '[SponsorBlock]: Interaction Reminder', 'c'])
        self._remove_marked_arrange_sponsors_test_impl(
            chapters, expected, [self._chapter(30, 40, remove=True)])

    def test_remove_marked_arrange_sponsors_ChapterWithSponsorCutInTheMiddle(self):
        cuts = [self._sponsor_chapter(20, 30, 'selfpromo', remove=True),
                self._chapter(40, 50, remove=True)]
        chapters = [
            *self._chapters([70], ['c']),
            self._sponsor_chapter(10, 60, 'sponsor'),
            *cuts]
        expected = self._chapters(
            [10, 40, 50], ['c', '[SponsorBlock]: Sponsor', 'c'])
        self._remove_marked_arrange_sponsors_test_impl(chapters, expected, cuts)

    def test_remove_marked_arrange_sponsors_ChapterWithCutHidingSponsor(self):
        cuts = [self._sponsor_chapter(20, 50, 'selfpromo', remove=True)]
        chapters = [
            *self._chapters([60], ['c']),
            self._sponsor_chapter(10, 20, 'intro'),
            self._sponsor_chapter(30, 40, 'sponsor'),
            self._sponsor_chapter(50, 60, 'outro'),
            *cuts]
        expected = self._chapters(
            [10, 20, 30], ['c', '[SponsorBlock]: Intermission/Intro Animation', '[SponsorBlock]: Endcards/Credits'])
        self._remove_marked_arrange_sponsors_test_impl(chapters, expected, cuts)

    def test_remove_marked_arrange_sponsors_ChapterWithAdjacentSponsors(self):
        chapters = [
            *self._chapters([70], ['c']),
            self._sponsor_chapter(10, 20, 'sponsor'),
            self._sponsor_chapter(20, 30, 'selfpromo'),
            self._sponsor_chapter(30, 40, 'interaction')]
        expected = self._chapters(
            [10, 20, 30, 40, 70],
            ['c', '[SponsorBlock]: Sponsor', '[SponsorBlock]: Unpaid/Self Promotion',
             '[SponsorBlock]: Interaction Reminder', 'c'])
        self._remove_marked_arrange_sponsors_test_impl(chapters, expected, [])

    def test_remove_marked_arrange_sponsors_ChapterWithAdjacentCuts(self):
        chapters = [
            *self._chapters([70], ['c']),
            self._sponsor_chapter(10, 20, 'sponsor'),
            self._sponsor_chapter(20, 30, 'interaction', remove=True),
            self._chapter(30, 40, remove=True),
            self._sponsor_chapter(40, 50, 'selfpromo', remove=True),
            self._sponsor_chapter(50, 60, 'interaction')]
        expected = self._chapters([10, 20, 30, 40],
                                  ['c', '[SponsorBlock]: Sponsor',
                                   '[SponsorBlock]: Interaction Reminder', 'c'])
        self._remove_marked_arrange_sponsors_test_impl(
            chapters, expected, [self._chapter(20, 50, remove=True)])

    def test_remove_marked_arrange_sponsors_ChapterWithOverlappingSponsors(self):
        chapters = [
            *self._chapters([70], ['c']),
            self._sponsor_chapter(10, 30, 'sponsor'),
            self._sponsor_chapter(20, 50, 'selfpromo'),
            self._sponsor_chapter(40, 60, 'interaction')]
        expected = self._chapters(
            [10, 20, 30, 40, 50, 60, 70],
            ['c', '[SponsorBlock]: Sponsor', '[SponsorBlock]: Sponsor, Unpaid/Self Promotion',
             '[SponsorBlock]: Unpaid/Self Promotion', '[SponsorBlock]: Unpaid/Self Promotion, Interaction Reminder',
             '[SponsorBlock]: Interaction Reminder', 'c'])
        self._remove_marked_arrange_sponsors_test_impl(chapters, expected, [])

    def test_remove_marked_arrange_sponsors_ChapterWithOverlappingCuts(self):
        chapters = [
            *self._chapters([70], ['c']),
            self._sponsor_chapter(10, 30, 'sponsor', remove=True),
            self._sponsor_chapter(20, 50, 'selfpromo', remove=True),
            self._sponsor_chapter(40, 60, 'interaction', remove=True)]
        self._remove_marked_arrange_sponsors_test_impl(
            chapters, self._chapters([20], ['c']), [self._chapter(10, 60, remove=True)])

    def test_remove_marked_arrange_sponsors_ChapterWithRunsOfOverlappingSponsors(self):
        chapters = [
            *self._chapters([170], ['c']),
            self._sponsor_chapter(0, 30, 'intro'),
            self._sponsor_chapter(20, 50, 'sponsor'),
            self._sponsor_chapter(40, 60, 'selfpromo'),
            self._sponsor_chapter(70, 90, 'sponsor'),
            self._sponsor_chapter(80, 100, 'sponsor'),
            self._sponsor_chapter(90, 110, 'sponsor'),
            self._sponsor_chapter(120, 140, 'selfpromo'),
            self._sponsor_chapter(130, 160, 'interaction'),
            self._sponsor_chapter(150, 170, 'outro')]
        expected = self._chapters(
            [20, 30, 40, 50, 60, 70, 110, 120, 130, 140, 150, 160, 170],
            ['[SponsorBlock]: Intermission/Intro Animation', '[SponsorBlock]: Intermission/Intro Animation, Sponsor', '[SponsorBlock]: Sponsor',
             '[SponsorBlock]: Sponsor, Unpaid/Self Promotion', '[SponsorBlock]: Unpaid/Self Promotion', 'c',
             '[SponsorBlock]: Sponsor', 'c', '[SponsorBlock]: Unpaid/Self Promotion',
             '[SponsorBlock]: Unpaid/Self Promotion, Interaction Reminder',
             '[SponsorBlock]: Interaction Reminder',
             '[SponsorBlock]: Interaction Reminder, Endcards/Credits', '[SponsorBlock]: Endcards/Credits'])
        self._remove_marked_arrange_sponsors_test_impl(chapters, expected, [])

    def test_remove_marked_arrange_sponsors_ChapterWithRunsOfOverlappingCuts(self):
        chapters = [
            *self._chapters([170], ['c']),
            self._chapter(0, 30, remove=True),
            self._sponsor_chapter(20, 50, 'sponsor', remove=True),
            self._chapter(40, 60, remove=True),
            self._sponsor_chapter(70, 90, 'sponsor', remove=True),
            self._chapter(80, 100, remove=True),
            self._chapter(90, 110, remove=True),
            self._sponsor_chapter(120, 140, 'sponsor', remove=True),
            self._sponsor_chapter(130, 160, 'selfpromo', remove=True),
            self._chapter(150, 170, remove=True)]
        expected_cuts = [self._chapter(0, 60, remove=True),
                         self._chapter(70, 110, remove=True),
                         self._chapter(120, 170, remove=True)]
        self._remove_marked_arrange_sponsors_test_impl(
            chapters, self._chapters([20], ['c']), expected_cuts)

    def test_remove_marked_arrange_sponsors_OverlappingSponsorsDifferentTitlesAfterCut(self):
        chapters = [
            *self._chapters([60], ['c']),
            self._sponsor_chapter(10, 60, 'sponsor'),
            self._sponsor_chapter(10, 40, 'intro'),
            self._sponsor_chapter(30, 50, 'interaction'),
            self._sponsor_chapter(30, 50, 'selfpromo', remove=True),
            self._sponsor_chapter(40, 50, 'interaction'),
            self._sponsor_chapter(50, 60, 'outro')]
        expected = self._chapters(
            [10, 30, 40], ['c', '[SponsorBlock]: Sponsor, Intermission/Intro Animation', '[SponsorBlock]: Sponsor, Endcards/Credits'])
        self._remove_marked_arrange_sponsors_test_impl(
            chapters, expected, [self._chapter(30, 50, remove=True)])

    def test_remove_marked_arrange_sponsors_SponsorsNoLongerOverlapAfterCut(self):
        chapters = [
            *self._chapters([70], ['c']),
            self._sponsor_chapter(10, 30, 'sponsor'),
            self._sponsor_chapter(20, 50, 'interaction'),
            self._sponsor_chapter(30, 50, 'selfpromo', remove=True),
            self._sponsor_chapter(40, 60, 'sponsor'),
            self._sponsor_chapter(50, 60, 'interaction')]
        expected = self._chapters(
            [10, 20, 40, 50], ['c', '[SponsorBlock]: Sponsor',
                               '[SponsorBlock]: Sponsor, Interaction Reminder', 'c'])
        self._remove_marked_arrange_sponsors_test_impl(
            chapters, expected, [self._chapter(30, 50, remove=True)])

    def test_remove_marked_arrange_sponsors_SponsorsStillOverlapAfterCut(self):
        chapters = [
            *self._chapters([70], ['c']),
            self._sponsor_chapter(10, 60, 'sponsor'),
            self._sponsor_chapter(20, 60, 'interaction'),
            self._sponsor_chapter(30, 50, 'selfpromo', remove=True)]
        expected = self._chapters(
            [10, 20, 40, 50], ['c', '[SponsorBlock]: Sponsor',
                               '[SponsorBlock]: Sponsor, Interaction Reminder', 'c'])
        self._remove_marked_arrange_sponsors_test_impl(
            chapters, expected, [self._chapter(30, 50, remove=True)])

    def test_remove_marked_arrange_sponsors_ChapterWithRunsOfOverlappingSponsorsAndCuts(self):
        chapters = [
            *self._chapters([200], ['c']),
            self._sponsor_chapter(10, 40, 'sponsor'),
            self._sponsor_chapter(10, 30, 'intro'),
            self._chapter(20, 30, remove=True),
            self._sponsor_chapter(30, 40, 'selfpromo'),
            self._sponsor_chapter(50, 70, 'sponsor'),
            self._sponsor_chapter(60, 80, 'interaction'),
            self._chapter(70, 80, remove=True),
            self._sponsor_chapter(70, 90, 'sponsor'),
            self._sponsor_chapter(80, 100, 'interaction'),
            self._sponsor_chapter(120, 170, 'selfpromo'),
            self._sponsor_chapter(130, 180, 'outro'),
            self._chapter(140, 150, remove=True),
            self._chapter(150, 160, remove=True)]
        expected = self._chapters(
            [10, 20, 30, 40, 50, 70, 80, 100, 110, 130, 140, 160],
            ['c', '[SponsorBlock]: Sponsor, Intermission/Intro Animation', '[SponsorBlock]: Sponsor, Unpaid/Self Promotion',
             'c', '[SponsorBlock]: Sponsor', '[SponsorBlock]: Sponsor, Interaction Reminder',
             '[SponsorBlock]: Interaction Reminder', 'c', '[SponsorBlock]: Unpaid/Self Promotion',
             '[SponsorBlock]: Unpaid/Self Promotion, Endcards/Credits', '[SponsorBlock]: Endcards/Credits', 'c'])
        expected_cuts = [self._chapter(20, 30, remove=True),
                         self._chapter(70, 80, remove=True),
                         self._chapter(140, 160, remove=True)]
        self._remove_marked_arrange_sponsors_test_impl(chapters, expected, expected_cuts)

    def test_remove_marked_arrange_sponsors_SponsorOverlapsMultipleChapters(self):
        chapters = [
            *self._chapters([20, 40, 60, 80, 100], ['c1', 'c2', 'c3', 'c4', 'c5']),
            self._sponsor_chapter(10, 90, 'sponsor')]
        expected = self._chapters([10, 90, 100], ['c1', '[SponsorBlock]: Sponsor', 'c5'])
        self._remove_marked_arrange_sponsors_test_impl(chapters, expected, [])

    def test_remove_marked_arrange_sponsors_CutOverlapsMultipleChapters(self):
        cuts = [self._chapter(10, 90, remove=True)]
        chapters = self._chapters([20, 40, 60, 80, 100], ['c1', 'c2', 'c3', 'c4', 'c5']) + cuts
        expected = self._chapters([10, 20], ['c1', 'c5'])
        self._remove_marked_arrange_sponsors_test_impl(chapters, expected, cuts)

    def test_remove_marked_arrange_sponsors_SponsorsWithinSomeChaptersAndOverlappingOthers(self):
        chapters = [
            *self._chapters([10, 40, 60, 80], ['c1', 'c2', 'c3', 'c4']),
            self._sponsor_chapter(20, 30, 'sponsor'),
            self._sponsor_chapter(50, 70, 'selfpromo')]
        expected = self._chapters([10, 20, 30, 40, 50, 70, 80],
                                  ['c1', 'c2', '[SponsorBlock]: Sponsor', 'c2', 'c3',
                                   '[SponsorBlock]: Unpaid/Self Promotion', 'c4'])
        self._remove_marked_arrange_sponsors_test_impl(chapters, expected, [])

    def test_remove_marked_arrange_sponsors_CutsWithinSomeChaptersAndOverlappingOthers(self):
        cuts = [self._chapter(20, 30, remove=True), self._chapter(50, 70, remove=True)]
        chapters = self._chapters([10, 40, 60, 80], ['c1', 'c2', 'c3', 'c4']) + cuts
        expected = self._chapters([10, 30, 40, 50], ['c1', 'c2', 'c3', 'c4'])
        self._remove_marked_arrange_sponsors_test_impl(chapters, expected, cuts)

    def test_remove_marked_arrange_sponsors_ChaptersAfterLastSponsor(self):
        chapters = [
            *self._chapters([20, 40, 50, 60], ['c1', 'c2', 'c3', 'c4']),
            self._sponsor_chapter(10, 30, 'music_offtopic')]
        expected = self._chapters(
            [10, 30, 40, 50, 60],
            ['c1', '[SponsorBlock]: Non-Music Section', 'c2', 'c3', 'c4'])
        self._remove_marked_arrange_sponsors_test_impl(chapters, expected, [])

    def test_remove_marked_arrange_sponsors_ChaptersAfterLastCut(self):
        cuts = [self._chapter(10, 30, remove=True)]
        chapters = self._chapters([20, 40, 50, 60], ['c1', 'c2', 'c3', 'c4']) + cuts
        expected = self._chapters([10, 20, 30, 40], ['c1', 'c2', 'c3', 'c4'])
        self._remove_marked_arrange_sponsors_test_impl(chapters, expected, cuts)

    def test_remove_marked_arrange_sponsors_SponsorStartsAtChapterStart(self):
        chapters = [
            *self._chapters([10, 20, 40], ['c1', 'c2', 'c3']),
            self._sponsor_chapter(20, 30, 'sponsor')]
        expected = self._chapters([10, 20, 30, 40], ['c1', 'c2', '[SponsorBlock]: Sponsor', 'c3'])
        self._remove_marked_arrange_sponsors_test_impl(chapters, expected, [])

    def test_remove_marked_arrange_sponsors_CutStartsAtChapterStart(self):
        cuts = [self._chapter(20, 30, remove=True)]
        chapters = self._chapters([10, 20, 40], ['c1', 'c2', 'c3']) + cuts
        expected = self._chapters([10, 20, 30], ['c1', 'c2', 'c3'])
        self._remove_marked_arrange_sponsors_test_impl(chapters, expected, cuts)

    def test_remove_marked_arrange_sponsors_SponsorEndsAtChapterEnd(self):
        chapters = [
            *self._chapters([10, 30, 40], ['c1', 'c2', 'c3']),
            self._sponsor_chapter(20, 30, 'sponsor')]
        expected = self._chapters([10, 20, 30, 40], ['c1', 'c2', '[SponsorBlock]: Sponsor', 'c3'])
        self._remove_marked_arrange_sponsors_test_impl(chapters, expected, [])

    def test_remove_marked_arrange_sponsors_CutEndsAtChapterEnd(self):
        cuts = [self._chapter(20, 30, remove=True)]
        chapters = self._chapters([10, 30, 40], ['c1', 'c2', 'c3']) + cuts
        expected = self._chapters([10, 20, 30], ['c1', 'c2', 'c3'])
        self._remove_marked_arrange_sponsors_test_impl(chapters, expected, cuts)

    def test_remove_marked_arrange_sponsors_SponsorCoincidesWithChapters(self):
        chapters = [
            *self._chapters([10, 20, 30, 40], ['c1', 'c2', 'c3', 'c4']),
            self._sponsor_chapter(10, 30, 'sponsor')]
        expected = self._chapters([10, 30, 40], ['c1', '[SponsorBlock]: Sponsor', 'c4'])
        self._remove_marked_arrange_sponsors_test_impl(chapters, expected, [])

    def test_remove_marked_arrange_sponsors_CutCoincidesWithChapters(self):
        cuts = [self._chapter(10, 30, remove=True)]
        chapters = self._chapters([10, 20, 30, 40], ['c1', 'c2', 'c3', 'c4']) + cuts
        expected = self._chapters([10, 20], ['c1', 'c4'])
        self._remove_marked_arrange_sponsors_test_impl(chapters, expected, cuts)

    def test_remove_marked_arrange_sponsors_SponsorsAtVideoBoundaries(self):
        chapters = [
            *self._chapters([20, 40, 60], ['c1', 'c2', 'c3']),
            self._sponsor_chapter(0, 10, 'intro'), self._sponsor_chapter(50, 60, 'outro')]
        expected = self._chapters(
            [10, 20, 40, 50, 60], ['[SponsorBlock]: Intermission/Intro Animation', 'c1', 'c2', 'c3', '[SponsorBlock]: Endcards/Credits'])
        self._remove_marked_arrange_sponsors_test_impl(chapters, expected, [])

    def test_remove_marked_arrange_sponsors_CutsAtVideoBoundaries(self):
        cuts = [self._chapter(0, 10, remove=True), self._chapter(50, 60, remove=True)]
        chapters = self._chapters([20, 40, 60], ['c1', 'c2', 'c3']) + cuts
        expected = self._chapters([10, 30, 40], ['c1', 'c2', 'c3'])
        self._remove_marked_arrange_sponsors_test_impl(chapters, expected, cuts)

    def test_remove_marked_arrange_sponsors_SponsorsOverlapChaptersAtVideoBoundaries(self):
        chapters = [
            *self._chapters([10, 40, 50], ['c1', 'c2', 'c3']),
            self._sponsor_chapter(0, 20, 'intro'),
            self._sponsor_chapter(30, 50, 'outro')]
        expected = self._chapters(
            [20, 30, 50], ['[SponsorBlock]: Intermission/Intro Animation', 'c2', '[SponsorBlock]: Endcards/Credits'])
        self._remove_marked_arrange_sponsors_test_impl(chapters, expected, [])

    def test_remove_marked_arrange_sponsors_CutsOverlapChaptersAtVideoBoundaries(self):
        cuts = [self._chapter(0, 20, remove=True), self._chapter(30, 50, remove=True)]
        chapters = self._chapters([10, 40, 50], ['c1', 'c2', 'c3']) + cuts
        expected = self._chapters([10], ['c2'])
        self._remove_marked_arrange_sponsors_test_impl(chapters, expected, cuts)

    def test_remove_marked_arrange_sponsors_EverythingSponsored(self):
        chapters = [
            *self._chapters([10, 20, 30, 40], ['c1', 'c2', 'c3', 'c4']),
            self._sponsor_chapter(0, 20, 'intro'),
            self._sponsor_chapter(20, 40, 'outro')]
        expected = self._chapters([20, 40], ['[SponsorBlock]: Intermission/Intro Animation', '[SponsorBlock]: Endcards/Credits'])
        self._remove_marked_arrange_sponsors_test_impl(chapters, expected, [])

    def test_remove_marked_arrange_sponsors_EverythingCut(self):
        cuts = [self._chapter(0, 20, remove=True), self._chapter(20, 40, remove=True)]
        chapters = self._chapters([10, 20, 30, 40], ['c1', 'c2', 'c3', 'c4']) + cuts
        self._remove_marked_arrange_sponsors_test_impl(
            chapters, [], [self._chapter(0, 40, remove=True)])

    def test_remove_marked_arrange_sponsors_TinyChaptersInTheOriginalArePreserved(self):
        chapters = self._chapters([0.1, 0.2, 0.3, 0.4], ['c1', 'c2', 'c3', 'c4'])
        self._remove_marked_arrange_sponsors_test_impl(chapters, chapters, [])

    def test_remove_marked_arrange_sponsors_TinySponsorsAreIgnored(self):
        chapters = [self._sponsor_chapter(0, 0.1, 'intro'), self._chapter(0.1, 0.2, 'c1'),
                    self._sponsor_chapter(0.2, 0.3, 'sponsor'), self._chapter(0.3, 0.4, 'c2'),
                    self._sponsor_chapter(0.4, 0.5, 'outro')]
        self._remove_marked_arrange_sponsors_test_impl(
            chapters, self._chapters([0.3, 0.5], ['c1', 'c2']), [])

    def test_remove_marked_arrange_sponsors_TinyChaptersResultingFromCutsAreIgnored(self):
        cuts = [self._chapter(1.5, 2.5, remove=True)]
        chapters = self._chapters([2, 3, 3.5], ['c1', 'c2', 'c3']) + cuts
        self._remove_marked_arrange_sponsors_test_impl(
            chapters, self._chapters([2, 2.5], ['c1', 'c3']), cuts)

    def test_remove_marked_arrange_sponsors_SingleTinyChapterIsPreserved(self):
        cuts = [self._chapter(0.5, 2, remove=True)]
        chapters = self._chapters([2], ['c']) + cuts
        self._remove_marked_arrange_sponsors_test_impl(
            chapters, self._chapters([0.5], ['c']), cuts)

    def test_remove_marked_arrange_sponsors_TinyChapterAtTheStartPrependedToTheNext(self):
        cuts = [self._chapter(0.5, 2, remove=True)]
        chapters = self._chapters([2, 4], ['c1', 'c2']) + cuts
        self._remove_marked_arrange_sponsors_test_impl(
            chapters, self._chapters([2.5], ['c2']), cuts)

    def test_remove_marked_arrange_sponsors_TinyChaptersResultingFromSponsorOverlapAreIgnored(self):
        chapters = [
            *self._chapters([1, 3, 4], ['c1', 'c2', 'c3']),
            self._sponsor_chapter(1.5, 2.5, 'sponsor')]
        self._remove_marked_arrange_sponsors_test_impl(
            chapters, self._chapters([1.5, 2.5, 4], ['c1', '[SponsorBlock]: Sponsor', 'c3']), [])

    def test_remove_marked_arrange_sponsors_TinySponsorsOverlapsAreIgnored(self):
        chapters = [
            *self._chapters([2, 3, 5], ['c1', 'c2', 'c3']),
            self._sponsor_chapter(1, 3, 'sponsor'),
            self._sponsor_chapter(2.5, 4, 'selfpromo')]
        self._remove_marked_arrange_sponsors_test_impl(
            chapters, self._chapters([1, 3, 4, 5], [
                'c1', '[SponsorBlock]: Sponsor', '[SponsorBlock]: Unpaid/Self Promotion', 'c3']), [])

    def test_remove_marked_arrange_sponsors_TinySponsorsPrependedToTheNextSponsor(self):
        chapters = [
            *self._chapters([4], ['c']),
            self._sponsor_chapter(1.5, 2, 'sponsor'),
            self._sponsor_chapter(2, 4, 'selfpromo')]
        self._remove_marked_arrange_sponsors_test_impl(
            chapters, self._chapters([1.5, 4], ['c', '[SponsorBlock]: Unpaid/Self Promotion']), [])

    def test_remove_marked_arrange_sponsors_SmallestSponsorInTheOverlapGetsNamed(self):
        self._pp._sponsorblock_chapter_title = '[SponsorBlock]: %(name)s'
        chapters = [
            *self._chapters([10], ['c']),
            self._sponsor_chapter(2, 8, 'sponsor'),
            self._sponsor_chapter(4, 6, 'selfpromo')]
        self._remove_marked_arrange_sponsors_test_impl(
            chapters, self._chapters([2, 4, 6, 8, 10], [
                'c', '[SponsorBlock]: Sponsor', '[SponsorBlock]: Unpaid/Self Promotion',
                '[SponsorBlock]: Sponsor', 'c',
            ]), [])

    def test_make_concat_opts_CommonCase(self):
        sponsor_chapters = [self._chapter(1, 2, 's1'), self._chapter(10, 20, 's2')]
        expected = '''ffconcat version 1.0
file 'file:test'
outpoint 1.000000
file 'file:test'
inpoint 2.000000
outpoint 10.000000
file 'file:test'
inpoint 20.000000
'''
        opts = self._pp._make_concat_opts(sponsor_chapters, 30)
        self.assertEqual(expected, ''.join(self._pp._concat_spec(['test'] * len(opts), opts)))

    def test_make_concat_opts_NoZeroDurationChunkAtVideoStart(self):
        sponsor_chapters = [self._chapter(0, 1, 's1'), self._chapter(10, 20, 's2')]
        expected = '''ffconcat version 1.0
file 'file:test'
inpoint 1.000000
outpoint 10.000000
file 'file:test'
inpoint 20.000000
'''
        opts = self._pp._make_concat_opts(sponsor_chapters, 30)
        self.assertEqual(expected, ''.join(self._pp._concat_spec(['test'] * len(opts), opts)))

    def test_make_concat_opts_NoZeroDurationChunkAtVideoEnd(self):
        sponsor_chapters = [self._chapter(1, 2, 's1'), self._chapter(10, 20, 's2')]
        expected = '''ffconcat version 1.0
file 'file:test'
outpoint 1.000000
file 'file:test'
inpoint 2.000000
outpoint 10.000000
'''
        opts = self._pp._make_concat_opts(sponsor_chapters, 20)
        self.assertEqual(expected, ''.join(self._pp._concat_spec(['test'] * len(opts), opts)))

    def test_quote_for_concat_RunsOfQuotes(self):
        self.assertEqual(
            r"'special '\'' '\'\''characters'\'\'\''galore'",
            self._pp._quote_for_ffmpeg("special ' ''characters'''galore"))

    def test_quote_for_concat_QuotesAtStart(self):
        self.assertEqual(
            r"\'\'\''special '\'' characters '\'' galore'",
            self._pp._quote_for_ffmpeg("'''special ' characters ' galore"))

    def test_quote_for_concat_QuotesAtEnd(self):
        self.assertEqual(
            r"'special '\'' characters '\'' galore'\'\'\'",
            self._pp._quote_for_ffmpeg("special ' characters ' galore'''"))


if __name__ == '__main__':
    unittest.main()



import pytest

from yt_dlp.extractor.youtube.pot._provider import IEContentProvider
from yt_dlp.cookies import YoutubeDLCookieJar
from yt_dlp.utils.networking import HTTPHeaderDict
from yt_dlp.extractor.youtube.pot.provider import (
    PoTokenRequest,
    PoTokenContext,
    ExternalRequestFeature,

)

from yt_dlp.extractor.youtube.pot.cache import (
    PoTokenCacheProvider,
    PoTokenCacheSpec,
    PoTokenCacheSpecProvider,
    CacheProviderWritePolicy,
)

import yt_dlp.extractor.youtube.pot.cache as cache

from yt_dlp.networking import Request
from yt_dlp.extractor.youtube.pot.provider import (
    PoTokenResponse,
    PoTokenProvider,
    PoTokenProviderRejectedRequest,
    provider_bug_report_message,
    register_provider,
    register_preference,
)

from yt_dlp.extractor.youtube.pot._registry import _pot_providers, _ptp_preferences, _pot_pcs_providers, _pot_cache_providers, _pot_cache_provider_preferences


class ExamplePTP(PoTokenProvider):
    PROVIDER_NAME = 'example'
    PROVIDER_VERSION = '0.0.1'
    BUG_REPORT_LOCATION = 'https://example.com/issues'

    _SUPPORTED_CLIENTS = ('WEB',)
    _SUPPORTED_CONTEXTS = (PoTokenContext.GVS, )

    _SUPPORTED_EXTERNAL_REQUEST_FEATURES = (
        ExternalRequestFeature.PROXY_SCHEME_HTTP,
        ExternalRequestFeature.PROXY_SCHEME_SOCKS5H,
    )

    def is_available(self) -> bool:
        return True

    def _real_request_pot(self, request: PoTokenRequest) -> PoTokenResponse:
        return PoTokenResponse('example-token', expires_at=123)


class ExampleCacheProviderPCP(PoTokenCacheProvider):

    PROVIDER_NAME = 'example'
    PROVIDER_VERSION = '0.0.1'
    BUG_REPORT_LOCATION = 'https://example.com/issues'

    def is_available(self) -> bool:
        return True

    def get(self, key: str):
        return 'example-cache'

    def store(self, key: str, value: str, expires_at: int):
        pass

    def delete(self, key: str):
        pass


class ExampleCacheSpecProviderPCSP(PoTokenCacheSpecProvider):

    PROVIDER_NAME = 'example'
    PROVIDER_VERSION = '0.0.1'
    BUG_REPORT_LOCATION = 'https://example.com/issues'

    def generate_cache_spec(self, request: PoTokenRequest):
        return PoTokenCacheSpec(
            key_bindings={'field': 'example-key'},
            default_ttl=60,
            write_policy=CacheProviderWritePolicy.WRITE_FIRST,
        )


class TestPoTokenProvider:

    def test_base_type(self):
        assert issubclass(PoTokenProvider, IEContentProvider)

    def test_create_provider_missing_fetch_method(self, ie, logger):
        class MissingMethodsPTP(PoTokenProvider):
            def is_available(self) -> bool:
                return True

        with pytest.raises(TypeError):
            MissingMethodsPTP(ie=ie, logger=logger, settings={})

    def test_create_provider_missing_available_method(self, ie, logger):
        class MissingMethodsPTP(PoTokenProvider):
            def _real_request_pot(self, request: PoTokenRequest) -> PoTokenResponse:
                raise PoTokenProviderRejectedRequest('Not implemented')

        with pytest.raises(TypeError):
            MissingMethodsPTP(ie=ie, logger=logger, settings={})

    def test_barebones_provider(self, ie, logger):
        class BarebonesProviderPTP(PoTokenProvider):
            def is_available(self) -> bool:
                return True

            def _real_request_pot(self, request: PoTokenRequest) -> PoTokenResponse:
                raise PoTokenProviderRejectedRequest('Not implemented')

        provider = BarebonesProviderPTP(ie=ie, logger=logger, settings={})
        assert provider.PROVIDER_NAME == 'BarebonesProvider'
        assert provider.PROVIDER_KEY == 'BarebonesProvider'
        assert provider.PROVIDER_VERSION == '0.0.0'
        assert provider.BUG_REPORT_MESSAGE == 'please report this issue to the provider developer at  (developer has not provided a bug report location)  .'

    def test_example_provider_success(self, ie, logger, pot_request):
        provider = ExamplePTP(ie=ie, logger=logger, settings={})
        assert provider.PROVIDER_NAME == 'example'
        assert provider.PROVIDER_KEY == 'Example'
        assert provider.PROVIDER_VERSION == '0.0.1'
        assert provider.BUG_REPORT_MESSAGE == 'please report this issue to the provider developer at  https://example.com/issues  .'
        assert provider.is_available()

        response = provider.request_pot(pot_request)

        assert response.po_token == 'example-token'
        assert response.expires_at == 123

    def test_provider_unsupported_context(self, ie, logger, pot_request):
        provider = ExamplePTP(ie=ie, logger=logger, settings={})
        pot_request.context = PoTokenContext.PLAYER

        with pytest.raises(PoTokenProviderRejectedRequest):
            provider.request_pot(pot_request)

    def test_provider_unsupported_client(self, ie, logger, pot_request):
        provider = ExamplePTP(ie=ie, logger=logger, settings={})
        pot_request.innertube_context['client']['clientName'] = 'ANDROID'

        with pytest.raises(PoTokenProviderRejectedRequest):
            provider.request_pot(pot_request)

    def test_provider_unsupported_proxy_scheme(self, ie, logger, pot_request):
        provider = ExamplePTP(ie=ie, logger=logger, settings={})
        pot_request.request_proxy = 'socks4://example.com'

        with pytest.raises(
            PoTokenProviderRejectedRequest,
            match=r'External requests by "example" provider do not support proxy scheme "socks4"\. Supported proxy '
            'schemes: http, socks5h',
        ):
            provider.request_pot(pot_request)

        pot_request.request_proxy = 'http://example.com'

        assert provider.request_pot(pot_request)

    def test_provider_ignore_external_request_features(self, ie, logger, pot_request):
        class InternalPTP(ExamplePTP):
            _SUPPORTED_EXTERNAL_REQUEST_FEATURES = None

        provider = InternalPTP(ie=ie, logger=logger, settings={})

        pot_request.request_proxy = 'socks5://example.com'
        assert provider.request_pot(pot_request)
        pot_request.request_source_address = '0.0.0.0'
        assert provider.request_pot(pot_request)

    def test_provider_unsupported_external_request_source_address(self, ie, logger, pot_request):
        class InternalPTP(ExamplePTP):
            _SUPPORTED_EXTERNAL_REQUEST_FEATURES = tuple()

        provider = InternalPTP(ie=ie, logger=logger, settings={})

        pot_request.request_source_address = None
        assert provider.request_pot(pot_request)

        pot_request.request_source_address = '0.0.0.0'
        with pytest.raises(
            PoTokenProviderRejectedRequest,
            match='External requests by "example" provider do not support setting source address',
        ):
            provider.request_pot(pot_request)

    def test_provider_supported_external_request_source_address(self, ie, logger, pot_request):
        class InternalPTP(ExamplePTP):
            _SUPPORTED_EXTERNAL_REQUEST_FEATURES = (
                ExternalRequestFeature.SOURCE_ADDRESS,
            )

        provider = InternalPTP(ie=ie, logger=logger, settings={})

        pot_request.request_source_address = None
        assert provider.request_pot(pot_request)

        pot_request.request_source_address = '0.0.0.0'
        assert provider.request_pot(pot_request)

    def test_provider_unsupported_external_request_tls_verification(self, ie, logger, pot_request):
        class InternalPTP(ExamplePTP):
            _SUPPORTED_EXTERNAL_REQUEST_FEATURES = tuple()

        provider = InternalPTP(ie=ie, logger=logger, settings={})

        pot_request.request_verify_tls = True
        assert provider.request_pot(pot_request)

        pot_request.request_verify_tls = False
        with pytest.raises(
            PoTokenProviderRejectedRequest,
            match='External requests by "example" provider do not support ignoring TLS certificate failures',
        ):
            provider.request_pot(pot_request)

    def test_provider_supported_external_request_tls_verification(self, ie, logger, pot_request):
        class InternalPTP(ExamplePTP):
            _SUPPORTED_EXTERNAL_REQUEST_FEATURES = (
                ExternalRequestFeature.DISABLE_TLS_VERIFICATION,
            )

        provider = InternalPTP(ie=ie, logger=logger, settings={})

        pot_request.request_verify_tls = True
        assert provider.request_pot(pot_request)

        pot_request.request_verify_tls = False
        assert provider.request_pot(pot_request)

    def test_provider_request_webpage(self, ie, logger, pot_request):
        provider = ExamplePTP(ie=ie, logger=logger, settings={})

        cookiejar = YoutubeDLCookieJar()
        pot_request.request_headers = HTTPHeaderDict({'User-Agent': 'example-user-agent'})
        pot_request.request_proxy = 'socks5://example-proxy.com'
        pot_request.request_cookiejar = cookiejar

        def mock_urlopen(request):
            return request

        ie._downloader.urlopen = mock_urlopen

        sent_request = provider._request_webpage(Request(
            'https://example.com',
        ), pot_request=pot_request)

        assert sent_request.url == 'https://example.com'
        assert sent_request.headers['User-Agent'] == 'example-user-agent'
        assert sent_request.proxies == {'all': 'socks5://example-proxy.com'}
        assert sent_request.extensions['cookiejar'] is cookiejar
        assert 'Requesting webpage' in logger.messages['info']

    def test_provider_request_webpage_override(self, ie, logger, pot_request):
        provider = ExamplePTP(ie=ie, logger=logger, settings={})

        cookiejar_request = YoutubeDLCookieJar()
        pot_request.request_headers = HTTPHeaderDict({'User-Agent': 'example-user-agent'})
        pot_request.request_proxy = 'socks5://example-proxy.com'
        pot_request.request_cookiejar = cookiejar_request

        def mock_urlopen(request):
            return request

        ie._downloader.urlopen = mock_urlopen

        sent_request = provider._request_webpage(Request(
            'https://example.com',
            headers={'User-Agent': 'override-user-agent-override'},
            proxies={'http': 'http://example-proxy-override.com'},
            extensions={'cookiejar': YoutubeDLCookieJar()},
        ), pot_request=pot_request, note='Custom requesting webpage')

        assert sent_request.url == 'https://example.com'
        assert sent_request.headers['User-Agent'] == 'override-user-agent-override'
        assert sent_request.proxies == {'http': 'http://example-proxy-override.com'}
        assert sent_request.extensions['cookiejar'] is not cookiejar_request
        assert 'Custom requesting webpage' in logger.messages['info']

    def test_provider_request_webpage_no_log(self, ie, logger, pot_request):
        provider = ExamplePTP(ie=ie, logger=logger, settings={})

        def mock_urlopen(request):
            return request

        ie._downloader.urlopen = mock_urlopen

        sent_request = provider._request_webpage(Request(
            'https://example.com',
        ), note=False)

        assert sent_request.url == 'https://example.com'
        assert 'info' not in logger.messages

    def test_provider_request_webpage_no_pot_request(self, ie, logger):
        provider = ExamplePTP(ie=ie, logger=logger, settings={})

        def mock_urlopen(request):
            return request

        ie._downloader.urlopen = mock_urlopen

        sent_request = provider._request_webpage(Request(
            'https://example.com',
        ), pot_request=None)

        assert sent_request.url == 'https://example.com'

    def test_get_config_arg(self, ie, logger):
        provider = ExamplePTP(ie=ie, logger=logger, settings={'abc': ['123D'], 'xyz': ['456a', '789B']})

        assert provider._configuration_arg('abc') == ['123d']
        assert provider._configuration_arg('abc', default=['default']) == ['123d']
        assert provider._configuration_arg('ABC', default=['default']) == ['default']
        assert provider._configuration_arg('abc', casesense=True) == ['123D']
        assert provider._configuration_arg('xyz', casesense=False) == ['456a', '789b']

    def test_require_class_end_with_suffix(self, ie, logger):
        class InvalidSuffix(PoTokenProvider):
            PROVIDER_NAME = 'invalid-suffix'

            def _real_request_pot(self, request: PoTokenRequest) -> PoTokenResponse:
                raise PoTokenProviderRejectedRequest('Not implemented')

            def is_available(self) -> bool:
                return True

        provider = InvalidSuffix(ie=ie, logger=logger, settings={})

        with pytest.raises(AssertionError):
            provider.PROVIDER_KEY  # noqa: B018


class TestPoTokenCacheProvider:

    def test_base_type(self):
        assert issubclass(PoTokenCacheProvider, IEContentProvider)

    def test_create_provider_missing_get_method(self, ie, logger):
        class MissingMethodsPCP(PoTokenCacheProvider):
            def store(self, key: str, value: str, expires_at: int):
                pass

            def delete(self, key: str):
                pass

                def is_available(self) -> bool:
                    return True

        with pytest.raises(TypeError):
            MissingMethodsPCP(ie=ie, logger=logger, settings={})

    def test_create_provider_missing_store_method(self, ie, logger):
        class MissingMethodsPCP(PoTokenCacheProvider):
            def get(self, key: str):
                pass

            def delete(self, key: str):
                pass

            def is_available(self) -> bool:
                return True

        with pytest.raises(TypeError):
            MissingMethodsPCP(ie=ie, logger=logger, settings={})

    def test_create_provider_missing_delete_method(self, ie, logger):
        class MissingMethodsPCP(PoTokenCacheProvider):
            def get(self, key: str):
                pass

            def store(self, key: str, value: str, expires_at: int):
                pass

            def is_available(self) -> bool:
                return True

        with pytest.raises(TypeError):
            MissingMethodsPCP(ie=ie, logger=logger, settings={})

    def test_create_provider_missing_is_available_method(self, ie, logger):
        class MissingMethodsPCP(PoTokenCacheProvider):
            def get(self, key: str):
                pass

            def store(self, key: str, value: str, expires_at: int):
                pass

            def delete(self, key: str):
                pass

        with pytest.raises(TypeError):
            MissingMethodsPCP(ie=ie, logger=logger, settings={})

    def test_barebones_provider(self, ie, logger):
        class BarebonesProviderPCP(PoTokenCacheProvider):

            def is_available(self) -> bool:
                return True

            def get(self, key: str):
                return 'example-cache'

            def store(self, key: str, value: str, expires_at: int):
                pass

            def delete(self, key: str):
                pass

        provider = BarebonesProviderPCP(ie=ie, logger=logger, settings={})
        assert provider.PROVIDER_NAME == 'BarebonesProvider'
        assert provider.PROVIDER_KEY == 'BarebonesProvider'
        assert provider.PROVIDER_VERSION == '0.0.0'
        assert provider.BUG_REPORT_MESSAGE == 'please report this issue to the provider developer at  (developer has not provided a bug report location)  .'

    def test_create_provider_example(self, ie, logger):
        provider = ExampleCacheProviderPCP(ie=ie, logger=logger, settings={})
        assert provider.PROVIDER_NAME == 'example'
        assert provider.PROVIDER_KEY == 'ExampleCacheProvider'
        assert provider.PROVIDER_VERSION == '0.0.1'
        assert provider.BUG_REPORT_MESSAGE == 'please report this issue to the provider developer at  https://example.com/issues  .'
        assert provider.is_available()

    def test_get_config_arg(self, ie, logger):
        provider = ExampleCacheProviderPCP(ie=ie, logger=logger, settings={'abc': ['123D'], 'xyz': ['456a', '789B']})
        assert provider._configuration_arg('abc') == ['123d']
        assert provider._configuration_arg('abc', default=['default']) == ['123d']
        assert provider._configuration_arg('ABC', default=['default']) == ['default']
        assert provider._configuration_arg('abc', casesense=True) == ['123D']
        assert provider._configuration_arg('xyz', casesense=False) == ['456a', '789b']

    def test_require_class_end_with_suffix(self, ie, logger):
        class InvalidSuffix(PoTokenCacheProvider):
            def get(self, key: str):
                return 'example-cache'

            def store(self, key: str, value: str, expires_at: int):
                pass

            def delete(self, key: str):
                pass

            def is_available(self) -> bool:
                return True

        provider = InvalidSuffix(ie=ie, logger=logger, settings={})

        with pytest.raises(AssertionError):
            provider.PROVIDER_KEY  # noqa: B018


class TestPoTokenCacheSpecProvider:

    def test_base_type(self):
        assert issubclass(PoTokenCacheSpecProvider, IEContentProvider)

    def test_create_provider_missing_supports_method(self, ie, logger):
        class MissingMethodsPCS(PoTokenCacheSpecProvider):
            pass

        with pytest.raises(TypeError):
            MissingMethodsPCS(ie=ie, logger=logger, settings={})

    def test_create_provider_barebones(self, ie, pot_request, logger):
        class BarebonesProviderPCSP(PoTokenCacheSpecProvider):
            def generate_cache_spec(self, request: PoTokenRequest):
                return PoTokenCacheSpec(
                    default_ttl=100,
                    key_bindings={},
                )

        provider = BarebonesProviderPCSP(ie=ie, logger=logger, settings={})
        assert provider.PROVIDER_NAME == 'BarebonesProvider'
        assert provider.PROVIDER_KEY == 'BarebonesProvider'
        assert provider.PROVIDER_VERSION == '0.0.0'
        assert provider.BUG_REPORT_MESSAGE == 'please report this issue to the provider developer at  (developer has not provided a bug report location)  .'
        assert provider.is_available()
        assert provider.generate_cache_spec(request=pot_request).default_ttl == 100
        assert provider.generate_cache_spec(request=pot_request).key_bindings == {}
        assert provider.generate_cache_spec(request=pot_request).write_policy == CacheProviderWritePolicy.WRITE_ALL

    def test_create_provider_example(self, ie, pot_request, logger):
        provider = ExampleCacheSpecProviderPCSP(ie=ie, logger=logger, settings={})
        assert provider.PROVIDER_NAME == 'example'
        assert provider.PROVIDER_KEY == 'ExampleCacheSpecProvider'
        assert provider.PROVIDER_VERSION == '0.0.1'
        assert provider.BUG_REPORT_MESSAGE == 'please report this issue to the provider developer at  https://example.com/issues  .'
        assert provider.is_available()
        assert provider.generate_cache_spec(pot_request)
        assert provider.generate_cache_spec(pot_request).key_bindings == {'field': 'example-key'}
        assert provider.generate_cache_spec(pot_request).default_ttl == 60
        assert provider.generate_cache_spec(pot_request).write_policy == CacheProviderWritePolicy.WRITE_FIRST

    def test_get_config_arg(self, ie, logger):
        provider = ExampleCacheSpecProviderPCSP(ie=ie, logger=logger, settings={'abc': ['123D'], 'xyz': ['456a', '789B']})

        assert provider._configuration_arg('abc') == ['123d']
        assert provider._configuration_arg('abc', default=['default']) == ['123d']
        assert provider._configuration_arg('ABC', default=['default']) == ['default']
        assert provider._configuration_arg('abc', casesense=True) == ['123D']
        assert provider._configuration_arg('xyz', casesense=False) == ['456a', '789b']

    def test_require_class_end_with_suffix(self, ie, logger):
        class InvalidSuffix(PoTokenCacheSpecProvider):
            def generate_cache_spec(self, request: PoTokenRequest):
                return None

        provider = InvalidSuffix(ie=ie, logger=logger, settings={})

        with pytest.raises(AssertionError):
            provider.PROVIDER_KEY  # noqa: B018


class TestPoTokenRequest:
    def test_copy_request(self, pot_request):
        copied_request = pot_request.copy()

        assert copied_request is not pot_request
        assert copied_request.context == pot_request.context
        assert copied_request.innertube_context == pot_request.innertube_context
        assert copied_request.innertube_context is not pot_request.innertube_context
        copied_request.innertube_context['client']['clientName'] = 'ANDROID'
        assert pot_request.innertube_context['client']['clientName'] != 'ANDROID'
        assert copied_request.innertube_host == pot_request.innertube_host
        assert copied_request.session_index == pot_request.session_index
        assert copied_request.player_url == pot_request.player_url
        assert copied_request.is_authenticated == pot_request.is_authenticated
        assert copied_request.visitor_data == pot_request.visitor_data
        assert copied_request.data_sync_id == pot_request.data_sync_id
        assert copied_request.video_id == pot_request.video_id
        assert copied_request.request_cookiejar is pot_request.request_cookiejar
        assert copied_request.request_proxy == pot_request.request_proxy
        assert copied_request.request_headers == pot_request.request_headers
        assert copied_request.request_headers is not pot_request.request_headers
        assert copied_request.request_timeout == pot_request.request_timeout
        assert copied_request.request_source_address == pot_request.request_source_address
        assert copied_request.request_verify_tls == pot_request.request_verify_tls
        assert copied_request.bypass_cache == pot_request.bypass_cache


def test_provider_bug_report_message(ie, logger):
    provider = ExamplePTP(ie=ie, logger=logger, settings={})
    assert provider.BUG_REPORT_MESSAGE == 'please report this issue to the provider developer at  https://example.com/issues  .'

    message = provider_bug_report_message(provider)
    assert message == '; please report this issue to the provider developer at  https://example.com/issues  .'

    message_before = provider_bug_report_message(provider, before='custom message!')
    assert message_before == 'custom message! Please report this issue to the provider developer at  https://example.com/issues  .'


def test_register_provider(ie):

    @register_provider
    class UnavailableProviderPTP(PoTokenProvider):
        def is_available(self) -> bool:
            return False

        def _real_request_pot(self, request: PoTokenRequest) -> PoTokenResponse:
            raise PoTokenProviderRejectedRequest('Not implemented')

    assert _pot_providers.value.get('UnavailableProvider') == UnavailableProviderPTP
    _pot_providers.value.pop('UnavailableProvider')


def test_register_pot_preference(ie):
    before = len(_ptp_preferences.value)

    @register_preference(ExamplePTP)
    def unavailable_preference(provider: PoTokenProvider, request: PoTokenRequest):
        return 1

    assert len(_ptp_preferences.value) == before + 1


def test_register_cache_provider(ie):

    @cache.register_provider
    class UnavailableCacheProviderPCP(PoTokenCacheProvider):
        def is_available(self) -> bool:
            return False

        def get(self, key: str):
            return 'example-cache'

        def store(self, key: str, value: str, expires_at: int):
            pass

        def delete(self, key: str):
            pass

    assert _pot_cache_providers.value.get('UnavailableCacheProvider') == UnavailableCacheProviderPCP
    _pot_cache_providers.value.pop('UnavailableCacheProvider')


def test_register_cache_provider_spec(ie):

    @cache.register_spec
    class UnavailableCacheProviderPCSP(PoTokenCacheSpecProvider):
        def is_available(self) -> bool:
            return False

        def generate_cache_spec(self, request: PoTokenRequest):
            return None

    assert _pot_pcs_providers.value.get('UnavailableCacheProvider') == UnavailableCacheProviderPCSP
    _pot_pcs_providers.value.pop('UnavailableCacheProvider')


def test_register_cache_provider_preference(ie):
    before = len(_pot_cache_provider_preferences.value)

    @cache.register_preference(ExampleCacheProviderPCP)
    def unavailable_preference(provider: PoTokenCacheProvider, request: PoTokenRequest):
        return 1

    assert len(_pot_cache_provider_preferences.value) == before + 1


def test_logger_log_level(logger):
    assert logger.LogLevel('INFO') == logger.LogLevel.INFO
    assert logger.LogLevel('debuG') == logger.LogLevel.DEBUG
    assert logger.LogLevel(10) == logger.LogLevel.DEBUG
    assert logger.LogLevel('UNKNOWN') == logger.LogLevel.INFO



import json

from .brightcove import BrightcoveNewIE
from .common import InfoExtractor
from .zype import ZypeIE
from ..networking import HEADRequest
from ..networking.exceptions import HTTPError
from ..utils import (
    ExtractorError,
    filter_dict,
    parse_qs,
    smuggle_url,
    try_call,
    urlencode_postdata,
)


class ThisOldHouseIE(InfoExtractor):
    _NETRC_MACHINE = 'thisoldhouse'
    _VALID_URL = r'https?://(?:www\.)?thisoldhouse\.com/(?:watch|how-to|tv-episode|(?:[^/?#]+/)?\d+)/(?P<id>[^/?#]+)'
    _TESTS = [{
        # Unresolved Brightcove URL embed (formerly Zype), free
        'url': 'https://www.thisoldhouse.com/furniture/21017078/how-to-build-a-storage-bench',
        'info_dict': {
            'id': '6325298523112',
            'ext': 'mp4',
            'title': 'How to Build a Storage Bench',
            'description': 'In the workshop, Tom Silva and Kevin O\'Connor build a storage bench for an entryway.',
            'timestamp': 1681793639,
            'upload_date': '20230418',
            'duration': 674.54,
            'tags': 'count:11',
            'uploader_id': '6314471934001',
            'thumbnail': r're:^https?://.*\.jpg',
        },
        'params': {
            'skip_download': True,
        },
    }, {
        # Brightcove embed, authwalled
        'url': 'https://www.thisoldhouse.com/glen-ridge-generational/99537/s45-e17-multi-generational',
        'info_dict': {
            'id': '6349675446112',
            'ext': 'mp4',
            'title': 'E17 | Glen Ridge Generational | Multi-Generational',
            'description': 'md5:53c6bc2e8031f3033d693d9a3563222c',
            'timestamp': 1711382202,
            'upload_date': '20240325',
            'duration': 1422.229,
            'tags': 'count:13',
            'uploader_id': '6314471934001',
            'thumbnail': r're:^https?://.*\.jpg',
        },
        'expected_warnings': ['Login with password is not supported for this website'],
        'params': {
            'skip_download': True,
        },
        'skip': 'Requires subscription',
    }, {
        # Page no longer has video
        'url': 'https://www.thisoldhouse.com/watch/arlington-arts-crafts-arts-and-crafts-class-begins',
        'only_matching': True,
    }, {
        # 404 Not Found
        'url': 'https://www.thisoldhouse.com/tv-episode/ask-toh-shelf-rough-electric',
        'only_matching': True,
    }, {
        # 404 Not Found
        'url': 'https://www.thisoldhouse.com/how-to/how-to-build-storage-bench',
        'only_matching': True,
    }, {
        'url': 'https://www.thisoldhouse.com/21113884/s41-e13-paradise-lost',
        'only_matching': True,
    }, {
        # iframe www.thisoldhouse.com
        'url': 'https://www.thisoldhouse.com/21083431/seaside-transformation-the-westerly-project',
        'only_matching': True,
    }]

    _LOGIN_URL = 'https://login.thisoldhouse.com/usernamepassword/login'

    def _perform_login(self, username, password):
        self._request_webpage(
            HEADRequest('https://www.thisoldhouse.com/insider'), None, 'Requesting session cookies')
        urlh = self._request_webpage(
            'https://www.thisoldhouse.com/wp-login.php', None, 'Requesting login info',
            errnote='Unable to login', query={'redirect_to': 'https://www.thisoldhouse.com/insider'})

        try:
            auth_form = self._download_webpage(
                self._LOGIN_URL, None, 'Submitting credentials', headers={
                    'Content-Type': 'application/json',
                    'Referer': urlh.url,
                }, data=json.dumps(filter_dict({
                    **{('client_id' if k == 'client' else k): v[0] for k, v in parse_qs(urlh.url).items()},
                    'tenant': 'thisoldhouse',
                    'username': username,
                    'password': password,
                    'popup_options': {},
                    'sso': True,
                    '_csrf': try_call(lambda: self._get_cookies(self._LOGIN_URL)['_csrf'].value),
                    '_intstate': 'deprecated',
                }), separators=(',', ':')).encode())
        except ExtractorError as e:
            if isinstance(e.cause, HTTPError) and e.cause.status == 401:
                raise ExtractorError('Invalid username or password', expected=True)
            raise

        self._request_webpage(
            'https://login.thisoldhouse.com/login/callback', None, 'Completing login',
            data=urlencode_postdata(self._hidden_inputs(auth_form)))

    def _real_extract(self, url):
        display_id = self._match_id(url)
        webpage = self._download_webpage(url, display_id)
        if 'To Unlock This content' in webpage:
            self.raise_login_required(
                'This video is only available for subscribers. '
                'Note that --cookies-from-browser may not work due to this site using session cookies')

        video_url, video_id = self._search_regex(
            r'<iframe[^>]+src=[\'"]((?:https?:)?//(?:www\.)?thisoldhouse\.(?:chorus\.build|com)/videos/zype/([0-9a-f]{24})[^\'"]*)[\'"]',
            webpage, 'zype url', group=(1, 2), default=(None, None))
        if video_url:
            video_url = self._request_webpage(HEADRequest(video_url), video_id, 'Resolving Zype URL').url
            return self.url_result(video_url, ZypeIE, video_id)

        video_url, video_id = self._search_regex([
            r'<iframe[^>]+src=[\'"]((?:https?:)?//players\.brightcove\.net/\d+/\w+/index\.html\?videoId=(\d+))',
            r'<iframe[^>]+src=[\'"]((?:https?:)?//(?:www\.)thisoldhouse\.com/videos/brightcove/(\d+))'],
            webpage, 'iframe url', group=(1, 2))
        if not parse_qs(video_url).get('videoId'):
            video_url = self._request_webpage(HEADRequest(video_url), video_id, 'Resolving Brightcove URL').url
        return self.url_result(smuggle_url(video_url, {'referrer': url}), BrightcoveNewIE, video_id)



from __future__ import annotations
import abc
import base64
import dataclasses
import hashlib
import json
import time
import pytest

from yt_dlp.extractor.youtube.pot._provider import BuiltinIEContentProvider, IEContentProvider

from yt_dlp.extractor.youtube.pot.provider import (
    PoTokenRequest,
    PoTokenContext,
    PoTokenProviderError,
    PoTokenProviderRejectedRequest,
)
from yt_dlp.extractor.youtube.pot._director import (
    PoTokenCache,
    validate_cache_spec,
    clean_pot,
    validate_response,
    PoTokenRequestDirector,
    provider_display_list,
)

from yt_dlp.extractor.youtube.pot.cache import (
    PoTokenCacheSpec,
    PoTokenCacheSpecProvider,
    PoTokenCacheProvider,
    CacheProviderWritePolicy,
    PoTokenCacheProviderError,
)


from yt_dlp.extractor.youtube.pot.provider import (
    PoTokenResponse,
    PoTokenProvider,
)


class BaseMockPoTokenProvider(PoTokenProvider, abc.ABC):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.available_called_times = 0
        self.request_called_times = 0
        self.close_called = False

    def is_available(self) -> bool:
        self.available_called_times += 1
        return True

    def request_pot(self, *args, **kwargs):
        self.request_called_times += 1
        return super().request_pot(*args, **kwargs)

    def close(self):
        self.close_called = True
        super().close()


class ExamplePTP(BaseMockPoTokenProvider):
    PROVIDER_NAME = 'example'
    PROVIDER_VERSION = '0.0.1'
    BUG_REPORT_LOCATION = 'https://example.com/issues'

    _SUPPORTED_CLIENTS = ('WEB',)
    _SUPPORTED_CONTEXTS = (PoTokenContext.GVS, )

    def _real_request_pot(self, request: PoTokenRequest) -> PoTokenResponse:
        if request.data_sync_id == 'example':
            return PoTokenResponse(request.video_id)
        return PoTokenResponse(EXAMPLE_PO_TOKEN)


def success_ptp(response: PoTokenResponse | None = None, key: str | None = None):
    class SuccessPTP(BaseMockPoTokenProvider):
        PROVIDER_NAME = 'success'
        PROVIDER_VERSION = '0.0.1'
        BUG_REPORT_LOCATION = 'https://success.example.com/issues'

        _SUPPORTED_CLIENTS = ('WEB',)
        _SUPPORTED_CONTEXTS = (PoTokenContext.GVS,)

        def _real_request_pot(self, request: PoTokenRequest) -> PoTokenResponse:
            return response or PoTokenResponse(EXAMPLE_PO_TOKEN)

    if key:
        SuccessPTP.PROVIDER_KEY = key
    return SuccessPTP


@pytest.fixture
def pot_provider(ie, logger):
    return success_ptp()(ie=ie, logger=logger, settings={})


class UnavailablePTP(BaseMockPoTokenProvider):
    PROVIDER_NAME = 'unavailable'
    BUG_REPORT_LOCATION = 'https://unavailable.example.com/issues'
    _SUPPORTED_CLIENTS = None
    _SUPPORTED_CONTEXTS = None

    def is_available(self) -> bool:
        super().is_available()
        return False

    def _real_request_pot(self, request: PoTokenRequest) -> PoTokenResponse:
        raise PoTokenProviderError('something went wrong')


class UnsupportedPTP(BaseMockPoTokenProvider):
    PROVIDER_NAME = 'unsupported'
    BUG_REPORT_LOCATION = 'https://unsupported.example.com/issues'
    _SUPPORTED_CLIENTS = None
    _SUPPORTED_CONTEXTS = None

    def _real_request_pot(self, request: PoTokenRequest) -> PoTokenResponse:
        raise PoTokenProviderRejectedRequest('unsupported request')


class ErrorPTP(BaseMockPoTokenProvider):
    PROVIDER_NAME = 'error'
    BUG_REPORT_LOCATION = 'https://error.example.com/issues'
    _SUPPORTED_CLIENTS = None
    _SUPPORTED_CONTEXTS = None

    def _real_request_pot(self, request: PoTokenRequest) -> PoTokenResponse:
        expected = request.video_id == 'expected'
        raise PoTokenProviderError('an error occurred', expected=expected)


class UnexpectedErrorPTP(BaseMockPoTokenProvider):
    PROVIDER_NAME = 'unexpected_error'
    BUG_REPORT_LOCATION = 'https://unexpected.example.com/issues'
    _SUPPORTED_CLIENTS = None
    _SUPPORTED_CONTEXTS = None

    def _real_request_pot(self, request: PoTokenRequest) -> PoTokenResponse:
        raise ValueError('an unexpected error occurred')


class InvalidPTP(BaseMockPoTokenProvider):
    PROVIDER_NAME = 'invalid'
    BUG_REPORT_LOCATION = 'https://invalid.example.com/issues'
    _SUPPORTED_CLIENTS = None
    _SUPPORTED_CONTEXTS = None

    def _real_request_pot(self, request: PoTokenRequest) -> PoTokenResponse:
        if request.video_id == 'invalid_type':
            return 'invalid-response'
        else:
            return PoTokenResponse('example-token?', expires_at='123')


class BaseMockCacheSpecProvider(PoTokenCacheSpecProvider, abc.ABC):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_called_times = 0
        self.is_available_called_times = 0
        self.close_called = False

    def is_available(self) -> bool:
        self.is_available_called_times += 1
        return super().is_available()

    def generate_cache_spec(self, request: PoTokenRequest):
        self.generate_called_times += 1

    def close(self):
        self.close_called = True
        super().close()


class ExampleCacheSpecProviderPCSP(BaseMockCacheSpecProvider):

    PROVIDER_NAME = 'example'
    PROVIDER_VERSION = '0.0.1'
    BUG_REPORT_LOCATION = 'https://example.com/issues'

    def generate_cache_spec(self, request: PoTokenRequest):
        super().generate_cache_spec(request)
        return PoTokenCacheSpec(
            key_bindings={'v': request.video_id, 'e': None},
            default_ttl=60,
        )


class UnavailableCacheSpecProviderPCSP(BaseMockCacheSpecProvider):

    PROVIDER_NAME = 'unavailable'
    PROVIDER_VERSION = '0.0.1'

    def is_available(self) -> bool:
        super().is_available()
        return False

    def generate_cache_spec(self, request: PoTokenRequest):
        super().generate_cache_spec(request)
        return None


class UnsupportedCacheSpecProviderPCSP(BaseMockCacheSpecProvider):

    PROVIDER_NAME = 'unsupported'
    PROVIDER_VERSION = '0.0.1'

    def generate_cache_spec(self, request: PoTokenRequest):
        super().generate_cache_spec(request)
        return None


class InvalidSpecCacheSpecProviderPCSP(BaseMockCacheSpecProvider):

    PROVIDER_NAME = 'invalid'
    PROVIDER_VERSION = '0.0.1'

    def generate_cache_spec(self, request: PoTokenRequest):
        super().generate_cache_spec(request)
        return 'invalid-spec'


class ErrorSpecCacheSpecProviderPCSP(BaseMockCacheSpecProvider):

    PROVIDER_NAME = 'invalid'
    PROVIDER_VERSION = '0.0.1'

    def generate_cache_spec(self, request: PoTokenRequest):
        super().generate_cache_spec(request)
        raise ValueError('something went wrong')


class BaseMockCacheProvider(PoTokenCacheProvider, abc.ABC):
    BUG_REPORT_MESSAGE = 'example bug report message'

    def __init__(self, *args, available=True, **kwargs):
        super().__init__(*args, **kwargs)
        self.store_calls = 0
        self.delete_calls = 0
        self.get_calls = 0
        self.available_called_times = 0
        self.available = available

    def is_available(self) -> bool:
        self.available_called_times += 1
        return self.available

    def store(self, *args, **kwargs):
        self.store_calls += 1

    def delete(self, *args, **kwargs):
        self.delete_calls += 1

    def get(self, *args, **kwargs):
        self.get_calls += 1

    def close(self):
        self.close_called = True
        super().close()


class ErrorPCP(BaseMockCacheProvider):
    PROVIDER_NAME = 'error'

    def store(self, *args, **kwargs):
        super().store(*args, **kwargs)
        raise PoTokenCacheProviderError('something went wrong')

    def get(self, *args, **kwargs):
        super().get(*args, **kwargs)
        raise PoTokenCacheProviderError('something went wrong')


class UnexpectedErrorPCP(BaseMockCacheProvider):
    PROVIDER_NAME = 'unexpected_error'

    def store(self, *args, **kwargs):
        super().store(*args, **kwargs)
        raise ValueError('something went wrong')

    def get(self, *args, **kwargs):
        super().get(*args, **kwargs)
        raise ValueError('something went wrong')


class MockMemoryPCP(BaseMockCacheProvider):
    PROVIDER_NAME = 'memory'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.cache = {}

    def store(self, key, value, expires_at):
        super().store(key, value, expires_at)
        self.cache[key] = (value, expires_at)

    def delete(self, key):
        super().delete(key)
        self.cache.pop(key, None)

    def get(self, key):
        super().get(key)
        return self.cache.get(key, [None])[0]


def create_memory_pcp(ie, logger, provider_key='memory', provider_name='memory', available=True):
    cache = MockMemoryPCP(ie, logger, {}, available=available)
    cache.PROVIDER_KEY = provider_key
    cache.PROVIDER_NAME = provider_name
    return cache


@pytest.fixture
def memorypcp(ie, logger) -> MockMemoryPCP:
    return create_memory_pcp(ie, logger)


@pytest.fixture
def pot_cache(ie, logger):
    class MockPoTokenCache(PoTokenCache):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.get_calls = 0
            self.store_calls = 0
            self.close_called = False

        def get(self, *args, **kwargs):
            self.get_calls += 1
            return super().get(*args, **kwargs)

        def store(self, *args, **kwargs):
            self.store_calls += 1
            return super().store(*args, **kwargs)

        def close(self):
            self.close_called = True
            super().close()

    return MockPoTokenCache(
        cache_providers=[MockMemoryPCP(ie, logger, {})],
        cache_spec_providers=[ExampleCacheSpecProviderPCSP(ie, logger, settings={})],
        logger=logger,
    )


EXAMPLE_PO_TOKEN = base64.urlsafe_b64encode(b'example-token').decode()


class TestPoTokenCache:

    def test_cache_success(self, memorypcp, pot_request, ie, logger):
        cache = PoTokenCache(
            cache_providers=[memorypcp],
            cache_spec_providers=[ExampleCacheSpecProviderPCSP(ie=ie, logger=logger, settings={})],
            logger=logger,
        )

        response = PoTokenResponse(EXAMPLE_PO_TOKEN)

        assert cache.get(pot_request) is None
        cache.store(pot_request, response)

        cached_response = cache.get(pot_request)
        assert cached_response is not None
        assert cached_response.po_token == EXAMPLE_PO_TOKEN
        assert cached_response.expires_at is not None

        assert cache.get(dataclasses.replace(pot_request, video_id='another-video-id')) is None

    def test_unsupported_cache_spec_no_fallback(self, memorypcp, pot_request, ie, logger):
        unsupported_provider = UnsupportedCacheSpecProviderPCSP(ie=ie, logger=logger, settings={})
        cache = PoTokenCache(
            cache_providers=[memorypcp],
            cache_spec_providers=[unsupported_provider],
            logger=logger,
        )

        response = PoTokenResponse(EXAMPLE_PO_TOKEN)
        assert cache.get(pot_request) is None
        assert unsupported_provider.generate_called_times == 1
        cache.store(pot_request, response)
        assert len(memorypcp.cache) == 0
        assert unsupported_provider.generate_called_times == 2
        assert cache.get(pot_request) is None
        assert unsupported_provider.generate_called_times == 3
        assert len(logger.messages.get('error', [])) == 0

    def test_unsupported_cache_spec_fallback(self, memorypcp, pot_request, ie, logger):
        unsupported_provider = UnsupportedCacheSpecProviderPCSP(ie=ie, logger=logger, settings={})
        example_provider = ExampleCacheSpecProviderPCSP(ie=ie, logger=logger, settings={})
        cache = PoTokenCache(
            cache_providers=[memorypcp],
            cache_spec_providers=[unsupported_provider, example_provider],
            logger=logger,
        )

        response = PoTokenResponse(EXAMPLE_PO_TOKEN)

        assert cache.get(pot_request) is None
        assert unsupported_provider.generate_called_times == 1
        assert example_provider.generate_called_times == 1

        cache.store(pot_request, response)
        assert unsupported_provider.generate_called_times == 2
        assert example_provider.generate_called_times == 2

        cached_response = cache.get(pot_request)
        assert unsupported_provider.generate_called_times == 3
        assert example_provider.generate_called_times == 3
        assert cached_response is not None
        assert cached_response.po_token == EXAMPLE_PO_TOKEN
        assert cached_response.expires_at is not None

        assert len(logger.messages.get('error', [])) == 0

    def test_invalid_cache_spec_no_fallback(self, memorypcp, pot_request, ie, logger):
        cache = PoTokenCache(
            cache_providers=[memorypcp],
            cache_spec_providers=[InvalidSpecCacheSpecProviderPCSP(ie=ie, logger=logger, settings={})],
            logger=logger,
        )

        response = PoTokenResponse(EXAMPLE_PO_TOKEN)

        assert cache.get(pot_request) is None
        cache.store(pot_request, response)

        assert cache.get(pot_request) is None

        assert 'PoTokenCacheSpecProvider "InvalidSpecCacheSpecProvider" generate_cache_spec() returned invalid spec invalid-spec; please report this issue to the provider developer at  (developer has not provided a bug report location)  .' in logger.messages['error']

    def test_invalid_cache_spec_fallback(self, memorypcp, pot_request, ie, logger):

        invalid_provider = InvalidSpecCacheSpecProviderPCSP(ie=ie, logger=logger, settings={})
        example_provider = ExampleCacheSpecProviderPCSP(ie=ie, logger=logger, settings={})
        cache = PoTokenCache(
            cache_providers=[memorypcp],
            cache_spec_providers=[invalid_provider, example_provider],
            logger=logger,
        )

        response = PoTokenResponse(EXAMPLE_PO_TOKEN)

        assert cache.get(pot_request) is None
        assert invalid_provider.generate_called_times == example_provider.generate_called_times == 1

        cache.store(pot_request, response)
        assert invalid_provider.generate_called_times == example_provider.generate_called_times == 2

        cached_response = cache.get(pot_request)
        assert invalid_provider.generate_called_times == example_provider.generate_called_times == 3
        assert cached_response is not None
        assert cached_response.po_token == EXAMPLE_PO_TOKEN
        assert cached_response.expires_at is not None

        assert 'PoTokenCacheSpecProvider "InvalidSpecCacheSpecProvider" generate_cache_spec() returned invalid spec invalid-spec; please report this issue to the provider developer at  (developer has not provided a bug report location)  .' in logger.messages['error']

    def test_unavailable_cache_spec_no_fallback(self, memorypcp, pot_request, ie, logger):
        unavailable_provider = UnavailableCacheSpecProviderPCSP(ie=ie, logger=logger, settings={})
        cache = PoTokenCache(
            cache_providers=[memorypcp],
            cache_spec_providers=[unavailable_provider],
            logger=logger,
        )

        response = PoTokenResponse(EXAMPLE_PO_TOKEN)

        assert cache.get(pot_request) is None
        cache.store(pot_request, response)
        assert cache.get(pot_request) is None
        assert unavailable_provider.generate_called_times == 0

    def test_unavailable_cache_spec_fallback(self, memorypcp, pot_request, ie, logger):
        unavailable_provider = UnavailableCacheSpecProviderPCSP(ie=ie, logger=logger, settings={})
        example_provider = ExampleCacheSpecProviderPCSP(ie=ie, logger=logger, settings={})
        cache = PoTokenCache(
            cache_providers=[memorypcp],
            cache_spec_providers=[unavailable_provider, example_provider],
            logger=logger,
        )

        response = PoTokenResponse(EXAMPLE_PO_TOKEN)

        assert cache.get(pot_request) is None
        assert unavailable_provider.generate_called_times == 0
        assert unavailable_provider.is_available_called_times == 1
        assert example_provider.generate_called_times == 1

        cache.store(pot_request, response)
        assert unavailable_provider.generate_called_times == 0
        assert unavailable_provider.is_available_called_times == 2
        assert example_provider.generate_called_times == 2

        cached_response = cache.get(pot_request)
        assert unavailable_provider.generate_called_times == 0
        assert unavailable_provider.is_available_called_times == 3
        assert example_provider.generate_called_times == 3
        assert example_provider.is_available_called_times == 3
        assert cached_response is not None
        assert cached_response.po_token == EXAMPLE_PO_TOKEN
        assert cached_response.expires_at is not None

    def test_unexpected_error_cache_spec(self, memorypcp, pot_request, ie, logger):
        error_provider = ErrorSpecCacheSpecProviderPCSP(ie=ie, logger=logger, settings={})
        cache = PoTokenCache(
            cache_providers=[memorypcp],
            cache_spec_providers=[error_provider],
            logger=logger,
        )

        response = PoTokenResponse(EXAMPLE_PO_TOKEN)

        assert cache.get(pot_request) is None
        cache.store(pot_request, response)
        assert cache.get(pot_request) is None
        assert error_provider.generate_called_times == 3
        assert error_provider.is_available_called_times == 3

        assert 'Error occurred with "invalid" PO Token cache spec provider: ValueError(\'something went wrong\'); please report this issue to the provider developer at  (developer has not provided a bug report location)  .' in logger.messages['error']

    def test_unexpected_error_cache_spec_fallback(self, memorypcp, pot_request, ie, logger):
        error_provider = ErrorSpecCacheSpecProviderPCSP(ie=ie, logger=logger, settings={})
        example_provider = ExampleCacheSpecProviderPCSP(ie=ie, logger=logger, settings={})
        cache = PoTokenCache(
            cache_providers=[memorypcp],
            cache_spec_providers=[error_provider, example_provider],
            logger=logger,
        )

        response = PoTokenResponse(EXAMPLE_PO_TOKEN)

        assert cache.get(pot_request) is None
        assert error_provider.generate_called_times == 1
        assert error_provider.is_available_called_times == 1
        assert example_provider.generate_called_times == 1

        cache.store(pot_request, response)
        assert error_provider.generate_called_times == 2
        assert error_provider.is_available_called_times == 2
        assert example_provider.generate_called_times == 2

        cached_response = cache.get(pot_request)
        assert error_provider.generate_called_times == 3
        assert error_provider.is_available_called_times == 3
        assert example_provider.generate_called_times == 3
        assert example_provider.is_available_called_times == 3
        assert cached_response is not None
        assert cached_response.po_token == EXAMPLE_PO_TOKEN
        assert cached_response.expires_at is not None

        assert 'Error occurred with "invalid" PO Token cache spec provider: ValueError(\'something went wrong\'); please report this issue to the provider developer at  (developer has not provided a bug report location)  .' in logger.messages['error']

    def test_key_bindings_spec_provider(self, memorypcp, pot_request, ie, logger):

        class ExampleProviderPCSP(PoTokenCacheSpecProvider):
            PROVIDER_NAME = 'example'

            def generate_cache_spec(self, request: PoTokenRequest):
                return PoTokenCacheSpec(
                    key_bindings={'v': request.video_id},
                    default_ttl=60,
                )

        class ExampleProviderTwoPCSP(ExampleProviderPCSP):
            pass

        example_provider = ExampleProviderPCSP(ie=ie, logger=logger, settings={})
        example_provider_two = ExampleProviderTwoPCSP(ie=ie, logger=logger, settings={})

        response = PoTokenResponse(EXAMPLE_PO_TOKEN)

        cache = PoTokenCache(
            cache_providers=[memorypcp],
            cache_spec_providers=[example_provider],
            logger=logger,
        )

        assert cache.get(pot_request) is None
        cache.store(pot_request, response)
        assert len(memorypcp.cache) == 1
        assert hashlib.sha256(
            f"{{'_dlp_cache': 'v1', '_p': 'ExampleProvider', 'v': '{pot_request.video_id}'}}".encode()).hexdigest() in memorypcp.cache

        # The second spec provider returns the exact same key bindings as the first one,
        # however the PoTokenCache should use the provider key to differentiate between them
        cache = PoTokenCache(
            cache_providers=[memorypcp],
            cache_spec_providers=[example_provider_two],
            logger=logger,
        )

        assert cache.get(pot_request) is None
        cache.store(pot_request, response)
        assert len(memorypcp.cache) == 2
        assert hashlib.sha256(
            f"{{'_dlp_cache': 'v1', '_p': 'ExampleProviderTwo', 'v': '{pot_request.video_id}'}}".encode()).hexdigest() in memorypcp.cache

    def test_cache_provider_preferences(self, pot_request, ie, logger):
        pcp_one = create_memory_pcp(ie, logger, provider_key='memory_pcp_one')
        pcp_two = create_memory_pcp(ie, logger, provider_key='memory_pcp_two')

        cache = PoTokenCache(
            cache_providers=[pcp_one, pcp_two],
            cache_spec_providers=[ExampleCacheSpecProviderPCSP(ie=ie, logger=logger, settings={})],
            logger=logger,
        )

        cache.store(pot_request, PoTokenResponse(EXAMPLE_PO_TOKEN), write_policy=CacheProviderWritePolicy.WRITE_FIRST)
        assert len(pcp_one.cache) == 1
        assert len(pcp_two.cache) == 0

        assert cache.get(pot_request)
        assert pcp_one.get_calls == 1
        assert pcp_two.get_calls == 0

        standard_preference_called = False
        pcp_one_preference_claled = False

        def standard_preference(provider, request, *_, **__):
            nonlocal standard_preference_called
            standard_preference_called = True
            assert isinstance(provider, PoTokenCacheProvider)
            assert isinstance(request, PoTokenRequest)
            return 1

        def pcp_one_preference(provider, request, *_, **__):
            nonlocal pcp_one_preference_claled
            pcp_one_preference_claled = True
            assert isinstance(provider, PoTokenCacheProvider)
            assert isinstance(request, PoTokenRequest)
            if provider.PROVIDER_KEY == pcp_one.PROVIDER_KEY:
                return -100
            return 0

        # test that it can hanldle multiple preferences
        cache.cache_provider_preferences.append(standard_preference)
        cache.cache_provider_preferences.append(pcp_one_preference)

        cache.store(pot_request, PoTokenResponse(EXAMPLE_PO_TOKEN), write_policy=CacheProviderWritePolicy.WRITE_FIRST)
        assert cache.get(pot_request)
        assert len(pcp_one.cache) == len(pcp_two.cache) == 1
        assert pcp_two.get_calls == pcp_one.get_calls == 1
        assert pcp_one.store_calls == pcp_two.store_calls == 1
        assert standard_preference_called
        assert pcp_one_preference_claled

    def test_secondary_cache_provider_hit(self, pot_request, ie, logger):
        pcp_one = create_memory_pcp(ie, logger, provider_key='memory_pcp_one')
        pcp_two = create_memory_pcp(ie, logger, provider_key='memory_pcp_two')

        cache = PoTokenCache(
            cache_providers=[pcp_two],
            cache_spec_providers=[ExampleCacheSpecProviderPCSP(ie=ie, logger=logger, settings={})],
            logger=logger,
        )

        # Given the lower priority provider has the cache hit, store the response in the higher priority provider
        cache.store(pot_request, PoTokenResponse(EXAMPLE_PO_TOKEN))
        assert cache.get(pot_request)

        cache.cache_providers[pcp_one.PROVIDER_KEY] = pcp_one

        def pcp_one_pref(provider, *_, **__):
            if provider.PROVIDER_KEY == pcp_one.PROVIDER_KEY:
                return 1
            return -1

        cache.cache_provider_preferences.append(pcp_one_pref)

        assert cache.get(pot_request)
        assert pcp_one.get_calls == 1
        assert pcp_two.get_calls == 2
        # Should write back to pcp_one (now the highest priority cache provider)
        assert pcp_one.store_calls == pcp_two.store_calls == 1
        assert 'Writing PO Token response to highest priority cache provider' in logger.messages['trace']

    def test_cache_provider_no_hits(self, pot_request, ie, logger):
        pcp_one = create_memory_pcp(ie, logger, provider_key='memory_pcp_one')
        pcp_two = create_memory_pcp(ie, logger, provider_key='memory_pcp_two')

        cache = PoTokenCache(
            cache_providers=[pcp_one, pcp_two],
            cache_spec_providers=[ExampleCacheSpecProviderPCSP(ie=ie, logger=logger, settings={})],
            logger=logger,
        )

        assert cache.get(pot_request) is None
        assert pcp_one.get_calls == pcp_two.get_calls == 1

    def test_get_invalid_po_token_response(self, pot_request, ie, logger):
        # Test various scenarios where the po token response stored in the cache provider is invalid
        pcp_one = create_memory_pcp(ie, logger, provider_key='memory_pcp_one')
        pcp_two = create_memory_pcp(ie, logger, provider_key='memory_pcp_two')

        cache = PoTokenCache(
            cache_providers=[pcp_one, pcp_two],
            cache_spec_providers=[ExampleCacheSpecProviderPCSP(ie=ie, logger=logger, settings={})],
            logger=logger,
        )

        valid_response = PoTokenResponse(EXAMPLE_PO_TOKEN)
        cache.store(pot_request, valid_response)
        assert len(pcp_one.cache) == len(pcp_two.cache) == 1
        # Overwrite the valid response with an invalid one in the cache
        pcp_one.store(next(iter(pcp_one.cache.keys())), json.dumps(dataclasses.asdict(PoTokenResponse(None))), int(time.time() + 1000))
        assert cache.get(pot_request).po_token == valid_response.po_token
        assert pcp_one.get_calls == pcp_two.get_calls == 1
        assert pcp_one.delete_calls == 1  # Invalid response should be deleted from cache
        assert pcp_one.store_calls == 3  # Since response was fetched from second cache provider, it should be stored in the first one
        assert len(pcp_one.cache) == 1
        assert 'Invalid PO Token response retrieved from cache provider "memory": {"po_token": null, "expires_at": null}; example bug report message' in logger.messages['error']

        # Overwrite the valid response with an invalid json in the cache
        pcp_one.store(next(iter(pcp_one.cache.keys())), 'invalid-json', int(time.time() + 1000))
        assert cache.get(pot_request).po_token == valid_response.po_token
        assert pcp_one.get_calls == pcp_two.get_calls == 2
        assert pcp_one.delete_calls == 2
        assert pcp_one.store_calls == 5  # 3 + 1 store we made in the test + 1 store from lower priority cache provider
        assert len(pcp_one.cache) == 1

        assert 'Invalid PO Token response retrieved from cache provider "memory": invalid-json; example bug report message' in logger.messages['error']

        # Valid json, but missing required fields
        pcp_one.store(next(iter(pcp_one.cache.keys())), '{"unknown_param": 0}', int(time.time() + 1000))
        assert cache.get(pot_request).po_token == valid_response.po_token
        assert pcp_one.get_calls == pcp_two.get_calls == 3
        assert pcp_one.delete_calls == 3
        assert pcp_one.store_calls == 7  # 5 + 1 store from test + 1 store from lower priority cache provider
        assert len(pcp_one.cache) == 1

        assert 'Invalid PO Token response retrieved from cache provider "memory": {"unknown_param": 0}; example bug report message' in logger.messages['error']

    def test_store_invalid_po_token_response(self, pot_request, ie, logger):
        # Should not store an invalid po token response
        pcp_one = create_memory_pcp(ie, logger, provider_key='memory_pcp_one')

        cache = PoTokenCache(
            cache_providers=[pcp_one],
            cache_spec_providers=[ExampleCacheSpecProviderPCSP(ie=ie, logger=logger, settings={})],
            logger=logger,
        )

        cache.store(pot_request, PoTokenResponse(po_token=EXAMPLE_PO_TOKEN, expires_at=80))
        assert cache.get(pot_request) is None
        assert pcp_one.store_calls == 0
        assert 'Invalid PO Token response provided to PoTokenCache.store()' in logger.messages['error'][0]

    def test_store_write_policy(self, pot_request, ie, logger):
        pcp_one = create_memory_pcp(ie, logger, provider_key='memory_pcp_one')
        pcp_two = create_memory_pcp(ie, logger, provider_key='memory_pcp_two')

        cache = PoTokenCache(
            cache_providers=[pcp_one, pcp_two],
            cache_spec_providers=[ExampleCacheSpecProviderPCSP(ie=ie, logger=logger, settings={})],
            logger=logger,
        )

        cache.store(pot_request, PoTokenResponse(EXAMPLE_PO_TOKEN), write_policy=CacheProviderWritePolicy.WRITE_FIRST)
        assert pcp_one.store_calls == 1
        assert pcp_two.store_calls == 0

        cache.store(pot_request, PoTokenResponse(EXAMPLE_PO_TOKEN), write_policy=CacheProviderWritePolicy.WRITE_ALL)
        assert pcp_one.store_calls == 2
        assert pcp_two.store_calls == 1

    def test_store_write_first_policy_cache_spec(self, pot_request, ie, logger):
        pcp_one = create_memory_pcp(ie, logger, provider_key='memory_pcp_one')
        pcp_two = create_memory_pcp(ie, logger, provider_key='memory_pcp_two')

        class WriteFirstPCSP(BaseMockCacheSpecProvider):
            def generate_cache_spec(self, request: PoTokenRequest):
                super().generate_cache_spec(request)
                return PoTokenCacheSpec(
                    key_bindings={'v': request.video_id, 'e': None},
                    default_ttl=60,
                    write_policy=CacheProviderWritePolicy.WRITE_FIRST,
                )

        cache = PoTokenCache(
            cache_providers=[pcp_one, pcp_two],
            cache_spec_providers=[WriteFirstPCSP(ie=ie, logger=logger, settings={})],
            logger=logger,
        )

        cache.store(pot_request, PoTokenResponse(EXAMPLE_PO_TOKEN))
        assert pcp_one.store_calls == 1
        assert pcp_two.store_calls == 0

    def test_store_write_all_policy_cache_spec(self, pot_request, ie, logger):
        pcp_one = create_memory_pcp(ie, logger, provider_key='memory_pcp_one')
        pcp_two = create_memory_pcp(ie, logger, provider_key='memory_pcp_two')

        class WriteAllPCSP(BaseMockCacheSpecProvider):
            def generate_cache_spec(self, request: PoTokenRequest):
                super().generate_cache_spec(request)
                return PoTokenCacheSpec(
                    key_bindings={'v': request.video_id, 'e': None},
                    default_ttl=60,
                    write_policy=CacheProviderWritePolicy.WRITE_ALL,
                )

        cache = PoTokenCache(
            cache_providers=[pcp_one, pcp_two],
            cache_spec_providers=[WriteAllPCSP(ie=ie, logger=logger, settings={})],
            logger=logger,
        )

        cache.store(pot_request, PoTokenResponse(EXAMPLE_PO_TOKEN))
        assert pcp_one.store_calls == 1
        assert pcp_two.store_calls == 1

    def test_expires_at_pot_response(self, pot_request, memorypcp, ie, logger):
        cache = PoTokenCache(
            cache_providers=[memorypcp],
            cache_spec_providers=[ExampleCacheSpecProviderPCSP(ie=ie, logger=logger, settings={})],
            logger=logger,
        )

        response = PoTokenResponse(EXAMPLE_PO_TOKEN, expires_at=10000000000)
        cache.store(pot_request, response)
        assert next(iter(memorypcp.cache.values()))[1] == 10000000000

    def test_expires_at_default_spec(self, pot_request, memorypcp, ie, logger):

        class TtlPCSP(BaseMockCacheSpecProvider):
            def generate_cache_spec(self, request: PoTokenRequest):
                super().generate_cache_spec(request)
                return PoTokenCacheSpec(
                    key_bindings={'v': request.video_id, 'e': None},
                    default_ttl=10000000000,
                )

        cache = PoTokenCache(
            cache_providers=[memorypcp],
            cache_spec_providers=[TtlPCSP(ie=ie, logger=logger, settings={})],
            logger=logger,
        )

        response = PoTokenResponse(EXAMPLE_PO_TOKEN)
        cache.store(pot_request, response)
        assert next(iter(memorypcp.cache.values()))[1] >= 10000000000

    def test_cache_provider_error_no_fallback(self, pot_request, ie, logger):
        error_pcp = ErrorPCP(ie, logger, {})
        cache = PoTokenCache(
            cache_providers=[error_pcp],
            cache_spec_providers=[ExampleCacheSpecProviderPCSP(ie=ie, logger=logger, settings={})],
            logger=logger,
        )

        response = PoTokenResponse(EXAMPLE_PO_TOKEN)
        cache.store(pot_request, response)
        assert cache.get(pot_request) is None
        assert error_pcp.get_calls == 1
        assert error_pcp.store_calls == 1

        assert logger.messages['warning'].count("Error from \"error\" PO Token cache provider: PoTokenCacheProviderError('something went wrong'); example bug report message") == 2

    def test_cache_provider_error_fallback(self, pot_request, ie, logger):
        error_pcp = ErrorPCP(ie, logger, {})
        memory_pcp = create_memory_pcp(ie, logger, provider_key='memory')

        cache = PoTokenCache(
            cache_providers=[error_pcp, memory_pcp],
            cache_spec_providers=[ExampleCacheSpecProviderPCSP(ie=ie, logger=logger, settings={})],
            logger=logger,
        )

        response = PoTokenResponse(EXAMPLE_PO_TOKEN)
        cache.store(pot_request, response)

        # 1. Store fails for error_pcp, stored in memory_pcp
        # 2. Get fails for error_pcp, fetched from memory_pcp
        # 3. Since fetched from lower priority, it should be stored in the highest priority cache provider
        # 4. Store fails in error_pcp. Since write policy is WRITE_FIRST, it should not try to store in memory_pcp regardless of if the store in error_pcp fails

        assert cache.get(pot_request)
        assert error_pcp.get_calls == 1
        assert error_pcp.store_calls == 2  # since highest priority, when fetched from lower priority, it should be stored in the highest priority cache provider
        assert memory_pcp.get_calls == 1
        assert memory_pcp.store_calls == 1

        assert logger.messages['warning'].count("Error from \"error\" PO Token cache provider: PoTokenCacheProviderError('something went wrong'); example bug report message") == 3

    def test_cache_provider_unexpected_error_no_fallback(self, pot_request, ie, logger):
        error_pcp = UnexpectedErrorPCP(ie, logger, {})
        cache = PoTokenCache(
            cache_providers=[error_pcp],
            cache_spec_providers=[ExampleCacheSpecProviderPCSP(ie=ie, logger=logger, settings={})],
            logger=logger,
        )

        response = PoTokenResponse(EXAMPLE_PO_TOKEN)
        cache.store(pot_request, response)
        assert cache.get(pot_request) is None
        assert error_pcp.get_calls == 1
        assert error_pcp.store_calls == 1

        assert logger.messages['error'].count("Error occurred with \"unexpected_error\" PO Token cache provider: ValueError('something went wrong'); example bug report message") == 2

    def test_cache_provider_unexpected_error_fallback(self, pot_request, ie, logger):
        error_pcp = UnexpectedErrorPCP(ie, logger, {})
        memory_pcp = create_memory_pcp(ie, logger, provider_key='memory')

        cache = PoTokenCache(
            cache_providers=[error_pcp, memory_pcp],
            cache_spec_providers=[ExampleCacheSpecProviderPCSP(ie=ie, logger=logger, settings={})],
            logger=logger,
        )

        response = PoTokenResponse(EXAMPLE_PO_TOKEN)
        cache.store(pot_request, response)

        # 1. Store fails for error_pcp, stored in memory_pcp
        # 2. Get fails for error_pcp, fetched from memory_pcp
        # 3. Since fetched from lower priority, it should be stored in the highest priority cache provider
        # 4. Store fails in error_pcp. Since write policy is WRITE_FIRST, it should not try to store in memory_pcp regardless of if the store in error_pcp fails

        assert cache.get(pot_request)
        assert error_pcp.get_calls == 1
        assert error_pcp.store_calls == 2  # since highest priority, when fetched from lower priority, it should be stored in the highest priority cache provider
        assert memory_pcp.get_calls == 1
        assert memory_pcp.store_calls == 1

        assert logger.messages['error'].count("Error occurred with \"unexpected_error\" PO Token cache provider: ValueError('something went wrong'); example bug report message") == 3

    def test_cache_provider_unavailable_no_fallback(self, pot_request, ie, logger):
        provider = create_memory_pcp(ie, logger, available=False)

        cache = PoTokenCache(
            cache_providers=[provider],
            cache_spec_providers=[ExampleCacheSpecProviderPCSP(ie=ie, logger=logger, settings={})],
            logger=logger,
        )

        response = PoTokenResponse(EXAMPLE_PO_TOKEN)
        cache.store(pot_request, response)
        assert cache.get(pot_request) is None
        assert provider.get_calls == 0
        assert provider.store_calls == 0
        assert provider.available_called_times

    def test_cache_provider_unavailable_fallback(self, pot_request, ie, logger):
        provider_unavailable = create_memory_pcp(ie, logger, provider_key='unavailable', provider_name='unavailable', available=False)
        provider_available = create_memory_pcp(ie, logger, provider_key='available', provider_name='available')

        cache = PoTokenCache(
            cache_providers=[provider_unavailable, provider_available],
            cache_spec_providers=[ExampleCacheSpecProviderPCSP(ie=ie, logger=logger, settings={})],
            logger=logger,
        )

        response = PoTokenResponse(EXAMPLE_PO_TOKEN)
        cache.store(pot_request, response)
        assert cache.get(pot_request) is not None
        assert provider_unavailable.get_calls == 0
        assert provider_unavailable.store_calls == 0
        assert provider_available.get_calls == 1
        assert provider_available.store_calls == 1
        assert provider_unavailable.available_called_times
        assert provider_available.available_called_times

        # should not even try to use the provider for the request
        assert 'Attempting to fetch a PO Token response from "unavailable" provider' not in logger.messages['trace']
        assert 'Attempting to fetch a PO Token response from "available" provider' not in logger.messages['trace']

    def test_available_not_called(self, ie, pot_request, logger):
        # Test that the available method is not called when provider higher in the list is available
        provider_unavailable = create_memory_pcp(
            ie, logger, provider_key='unavailable', provider_name='unavailable', available=False)
        provider_available = create_memory_pcp(ie, logger, provider_key='available', provider_name='available')

        logger.log_level = logger.LogLevel.INFO

        cache = PoTokenCache(
            cache_providers=[provider_available, provider_unavailable],
            cache_spec_providers=[ExampleCacheSpecProviderPCSP(ie=ie, logger=logger, settings={})],
            logger=logger,
        )

        response = PoTokenResponse(EXAMPLE_PO_TOKEN)
        cache.store(pot_request, response, write_policy=CacheProviderWritePolicy.WRITE_FIRST)
        assert cache.get(pot_request) is not None
        assert provider_unavailable.get_calls == 0
        assert provider_unavailable.store_calls == 0
        assert provider_available.get_calls == 1
        assert provider_available.store_calls == 1
        assert provider_unavailable.available_called_times == 0
        assert provider_available.available_called_times
        assert 'PO Token Cache Providers: available-0.0.0 (external), unavailable-0.0.0 (external, unavailable)' not in logger.messages.get('trace', [])

    def test_available_called_trace(self, ie, pot_request, logger):
        # But if logging level is trace should call available (as part of debug logging)
        provider_unavailable = create_memory_pcp(
            ie, logger, provider_key='unavailable', provider_name='unavailable', available=False)
        provider_available = create_memory_pcp(ie, logger, provider_key='available', provider_name='available')

        logger.log_level = logger.LogLevel.TRACE

        cache = PoTokenCache(
            cache_providers=[provider_available, provider_unavailable],
            cache_spec_providers=[ExampleCacheSpecProviderPCSP(ie=ie, logger=logger, settings={})],
            logger=logger,
        )

        response = PoTokenResponse(EXAMPLE_PO_TOKEN)
        cache.store(pot_request, response, write_policy=CacheProviderWritePolicy.WRITE_FIRST)
        assert cache.get(pot_request) is not None
        assert provider_unavailable.get_calls == 0
        assert provider_unavailable.store_calls == 0
        assert provider_available.get_calls == 1
        assert provider_available.store_calls == 1
        assert provider_unavailable.available_called_times
        assert provider_available.available_called_times
        assert 'PO Token Cache Providers: available-0.0.0 (external), unavailable-0.0.0 (external, unavailable)' in logger.messages.get('trace', [])

    def test_close(self, ie, pot_request, logger):
        # Should call close on the cache providers and cache specs
        memory_pcp = create_memory_pcp(ie, logger, provider_key='memory')
        memory2_pcp = create_memory_pcp(ie, logger, provider_key='memory2')

        spec1 = ExampleCacheSpecProviderPCSP(ie=ie, logger=logger, settings={})
        spec2 = UnavailableCacheSpecProviderPCSP(ie=ie, logger=logger, settings={})

        cache = PoTokenCache(
            cache_providers=[memory2_pcp, memory_pcp],
            cache_spec_providers=[spec1, spec2],
            logger=logger,
        )

        cache.close()
        assert memory_pcp.close_called
        assert memory2_pcp.close_called
        assert spec1.close_called
        assert spec2.close_called


class TestPoTokenRequestDirector:

    def test_request_pot_success(self, ie, pot_request, pot_cache, pot_provider, logger):
        director = PoTokenRequestDirector(logger=logger, cache=pot_cache)
        director.register_provider(pot_provider)
        response = director.get_po_token(pot_request)
        assert response == EXAMPLE_PO_TOKEN

    def test_request_and_cache(self, ie, pot_request, pot_cache, pot_provider, logger):
        director = PoTokenRequestDirector(logger=logger, cache=pot_cache)
        director.register_provider(pot_provider)
        response = director.get_po_token(pot_request)
        assert response == EXAMPLE_PO_TOKEN
        assert pot_provider.request_called_times == 1
        assert pot_cache.get_calls == 1
        assert pot_cache.store_calls == 1

        # Second request, should be cached
        response = director.get_po_token(pot_request)
        assert response == EXAMPLE_PO_TOKEN
        assert pot_cache.get_calls == 2
        assert pot_cache.store_calls == 1
        assert pot_provider.request_called_times == 1

    def test_bypass_cache(self, ie, pot_request, pot_cache, logger, pot_provider):
        pot_request.bypass_cache = True

        director = PoTokenRequestDirector(logger=logger, cache=pot_cache)
        director.register_provider(pot_provider)
        response = director.get_po_token(pot_request)
        assert response == EXAMPLE_PO_TOKEN
        assert pot_provider.request_called_times == 1
        assert pot_cache.get_calls == 0
        assert pot_cache.store_calls == 1

        # Second request, should not get from cache
        response = director.get_po_token(pot_request)
        assert response == EXAMPLE_PO_TOKEN
        assert pot_provider.request_called_times == 2
        assert pot_cache.get_calls == 0
        assert pot_cache.store_calls == 2

        # POT is still cached, should get from cache
        pot_request.bypass_cache = False
        response = director.get_po_token(pot_request)
        assert response == EXAMPLE_PO_TOKEN
        assert pot_provider.request_called_times == 2
        assert pot_cache.get_calls == 1
        assert pot_cache.store_calls == 2

    def test_clean_pot_generate(self, ie, pot_request, pot_cache, logger):
        # Token should be cleaned before returning
        base_token = base64.urlsafe_b64encode(b'token').decode()
        director = PoTokenRequestDirector(logger=logger, cache=pot_cache)
        provider = success_ptp(PoTokenResponse(base_token + '?extra=params'))(ie, logger, settings={})
        director.register_provider(provider)

        response = director.get_po_token(pot_request)
        assert response == base_token
        assert provider.request_called_times == 1

        # Confirm the cleaned version was stored in the cache
        cached_token = pot_cache.get(pot_request)
        assert cached_token.po_token == base_token

    def test_clean_pot_cache(self, ie, pot_request, pot_cache, logger, pot_provider):
        # Token retrieved from cache should be cleaned before returning
        base_token = base64.urlsafe_b64encode(b'token').decode()
        pot_cache.store(pot_request, PoTokenResponse(base_token + '?extra=params'))
        director = PoTokenRequestDirector(logger=logger, cache=pot_cache)
        director.register_provider(pot_provider)

        response = director.get_po_token(pot_request)
        assert response == base_token
        assert pot_cache.get_calls == 1
        assert pot_provider.request_called_times == 0

    def test_cache_expires_at_none(self, ie, pot_request, pot_cache, logger, pot_provider):
        # Should cache if expires_at=None in the response
        director = PoTokenRequestDirector(logger=logger, cache=pot_cache)
        provider = success_ptp(PoTokenResponse(EXAMPLE_PO_TOKEN, expires_at=None))(ie, logger, settings={})
        director.register_provider(provider)

        response = director.get_po_token(pot_request)
        assert response == EXAMPLE_PO_TOKEN
        assert pot_cache.store_calls == 1
        assert pot_cache.get(pot_request).po_token == EXAMPLE_PO_TOKEN

    def test_cache_expires_at_positive(self, ie, pot_request, pot_cache, logger, pot_provider):
        # Should cache if expires_at is a positive number in the response
        director = PoTokenRequestDirector(logger=logger, cache=pot_cache)
        provider = success_ptp(PoTokenResponse(EXAMPLE_PO_TOKEN, expires_at=99999999999))(ie, logger, settings={})
        director.register_provider(provider)

        response = director.get_po_token(pot_request)
        assert response == EXAMPLE_PO_TOKEN
        assert pot_cache.store_calls == 1
        assert pot_cache.get(pot_request).po_token == EXAMPLE_PO_TOKEN

    @pytest.mark.parametrize('expires_at', [0, -1])
    def test_not_cache_expires_at(self, ie, pot_request, pot_cache, logger, pot_provider, expires_at):
        # Should not cache if expires_at <= 0 in the response
        director = PoTokenRequestDirector(logger=logger, cache=pot_cache)
        provider = success_ptp(PoTokenResponse(EXAMPLE_PO_TOKEN, expires_at=expires_at))(ie, logger, settings={})
        director.register_provider(provider)

        response = director.get_po_token(pot_request)
        assert response == EXAMPLE_PO_TOKEN
        assert pot_cache.store_calls == 0
        assert pot_cache.get(pot_request) is None

    def test_no_providers(self, ie, pot_request, pot_cache, logger):
        director = PoTokenRequestDirector(logger=logger, cache=pot_cache)
        response = director.get_po_token(pot_request)
        assert response is None

    def test_try_cache_no_providers(self, ie, pot_request, pot_cache, logger):
        # Should still try the cache even if no providers are configured
        pot_cache.store(pot_request, PoTokenResponse(EXAMPLE_PO_TOKEN))
        director = PoTokenRequestDirector(logger=logger, cache=pot_cache)

        response = director.get_po_token(pot_request)
        assert response == EXAMPLE_PO_TOKEN

    def test_close(self, ie, pot_request, pot_cache, pot_provider, logger):
        # Should call close on the pot cache and any providers
        director = PoTokenRequestDirector(logger=logger, cache=pot_cache)

        provider2 = UnavailablePTP(ie, logger, {})
        director.register_provider(pot_provider)
        director.register_provider(provider2)

        director.close()
        assert pot_provider.close_called
        assert provider2.close_called
        assert pot_cache.close_called

    def test_pot_provider_preferences(self, pot_request, pot_cache, ie, logger):
        pot_request.bypass_cache = True
        provider_two_pot = base64.urlsafe_b64encode(b'token2').decode()

        example_provider = success_ptp(response=PoTokenResponse(EXAMPLE_PO_TOKEN), key='exampleone')(ie, logger, settings={})
        example_provider_two = success_ptp(response=PoTokenResponse(provider_two_pot), key='exampletwo')(ie, logger, settings={})

        director = PoTokenRequestDirector(logger=logger, cache=pot_cache)
        director.register_provider(example_provider)
        director.register_provider(example_provider_two)

        response = director.get_po_token(pot_request)
        assert response == EXAMPLE_PO_TOKEN
        assert example_provider.request_called_times == 1
        assert example_provider_two.request_called_times == 0

        standard_preference_called = False
        example_preference_called = False

        # Test that the provider preferences are respected
        def standard_preference(provider, request, *_, **__):
            nonlocal standard_preference_called
            standard_preference_called = True
            assert isinstance(provider, PoTokenProvider)
            assert isinstance(request, PoTokenRequest)
            return 1

        def example_preference(provider, request, *_, **__):
            nonlocal example_preference_called
            example_preference_called = True
            assert isinstance(provider, PoTokenProvider)
            assert isinstance(request, PoTokenRequest)
            if provider.PROVIDER_KEY == example_provider.PROVIDER_KEY:
                return -100
            return 0

        # test that it can handle multiple preferences
        director.register_preference(example_preference)
        director.register_preference(standard_preference)

        response = director.get_po_token(pot_request)
        assert response == provider_two_pot
        assert example_provider.request_called_times == 1
        assert example_provider_two.request_called_times == 1
        assert standard_preference_called
        assert example_preference_called

    def test_unsupported_request_no_fallback(self, ie, logger, pot_cache, pot_request):
        director = PoTokenRequestDirector(logger=logger, cache=pot_cache)
        provider = UnsupportedPTP(ie, logger, {})
        director.register_provider(provider)

        response = director.get_po_token(pot_request)
        assert response is None
        assert provider.request_called_times == 1

    def test_unsupported_request_fallback(self, ie, logger, pot_cache, pot_request, pot_provider):
        # Should fallback to the next provider if the first one does not support the request
        director = PoTokenRequestDirector(logger=logger, cache=pot_cache)
        provider = UnsupportedPTP(ie, logger, {})
        director.register_provider(provider)
        director.register_provider(pot_provider)

        response = director.get_po_token(pot_request)
        assert response == EXAMPLE_PO_TOKEN
        assert provider.request_called_times == 1
        assert pot_provider.request_called_times == 1
        assert 'PO Token Provider "unsupported" rejected this request, trying next available provider. Reason: unsupported request' in logger.messages['trace']

    def test_unavailable_request_no_fallback(self, ie, logger, pot_cache, pot_request):
        director = PoTokenRequestDirector(logger=logger, cache=pot_cache)
        provider = UnavailablePTP(ie, logger, {})
        director.register_provider(provider)

        response = director.get_po_token(pot_request)
        assert response is None
        assert provider.request_called_times == 0
        assert provider.available_called_times

    def test_unavailable_request_fallback(self, ie, logger, pot_cache, pot_request, pot_provider):
        # Should fallback to the next provider if the first one is unavailable
        director = PoTokenRequestDirector(logger=logger, cache=pot_cache)
        provider = UnavailablePTP(ie, logger, {})
        director.register_provider(provider)
        director.register_provider(pot_provider)

        response = director.get_po_token(pot_request)
        assert response == EXAMPLE_PO_TOKEN
        assert provider.request_called_times == 0
        assert provider.available_called_times
        assert pot_provider.request_called_times == 1
        assert pot_provider.available_called_times
        # should not even try use the provider for the request
        assert 'Attempting to fetch a PO Token from "unavailable" provider' not in logger.messages['trace']
        assert 'Attempting to fetch a PO Token from "success" provider' in logger.messages['trace']

    def test_available_not_called(self, ie, logger, pot_cache, pot_request, pot_provider):
        # Test that the available method is not called when provider higher in the list is available
        logger.log_level = logger.LogLevel.INFO
        director = PoTokenRequestDirector(logger=logger, cache=pot_cache)
        provider = UnavailablePTP(ie, logger, {})
        director.register_provider(pot_provider)
        director.register_provider(provider)

        response = director.get_po_token(pot_request)
        assert response == EXAMPLE_PO_TOKEN
        assert provider.request_called_times == 0
        assert provider.available_called_times == 0
        assert pot_provider.request_called_times == 1
        assert pot_provider.available_called_times == 2
        assert 'PO Token Providers: success-0.0.1 (external), unavailable-0.0.0 (external, unavailable)' not in logger.messages.get('trace', [])

    def test_available_called_trace(self, ie, logger, pot_cache, pot_request, pot_provider):
        # But if logging level is trace should call available (as part of debug logging)
        logger.log_level = logger.LogLevel.TRACE
        director = PoTokenRequestDirector(logger=logger, cache=pot_cache)
        provider = UnavailablePTP(ie, logger, {})
        director.register_provider(pot_provider)
        director.register_provider(provider)

        response = director.get_po_token(pot_request)
        assert response == EXAMPLE_PO_TOKEN
        assert provider.request_called_times == 0
        assert provider.available_called_times == 1
        assert pot_provider.request_called_times == 1
        assert pot_provider.available_called_times == 3
        assert 'PO Token Providers: success-0.0.1 (external), unavailable-0.0.0 (external, unavailable)' in logger.messages['trace']

    def test_provider_error_no_fallback_unexpected(self, ie, logger, pot_cache, pot_request):
        director = PoTokenRequestDirector(logger=logger, cache=pot_cache)
        provider = ErrorPTP(ie, logger, {})
        director.register_provider(provider)
        pot_request.video_id = 'unexpected'
        response = director.get_po_token(pot_request)
        assert response is None
        assert provider.request_called_times == 1
        assert "Error fetching PO Token from \"error\" provider: PoTokenProviderError('an error occurred'); please report this issue to the provider developer at  https://error.example.com/issues  ." in logger.messages['warning']

    def test_provider_error_no_fallback_expected(self, ie, logger, pot_cache, pot_request):
        director = PoTokenRequestDirector(logger=logger, cache=pot_cache)
        provider = ErrorPTP(ie, logger, {})
        director.register_provider(provider)
        pot_request.video_id = 'expected'
        response = director.get_po_token(pot_request)
        assert response is None
        assert provider.request_called_times == 1
        assert "Error fetching PO Token from \"error\" provider: PoTokenProviderError('an error occurred')" in logger.messages['warning']

    def test_provider_error_fallback(self, ie, logger, pot_cache, pot_request, pot_provider):
        # Should fallback to the next provider if the first one raises an error
        director = PoTokenRequestDirector(logger=logger, cache=pot_cache)
        provider = ErrorPTP(ie, logger, {})
        director.register_provider(provider)
        director.register_provider(pot_provider)

        response = director.get_po_token(pot_request)
        assert response == EXAMPLE_PO_TOKEN
        assert provider.request_called_times == 1
        assert pot_provider.request_called_times == 1
        assert "Error fetching PO Token from \"error\" provider: PoTokenProviderError('an error occurred'); please report this issue to the provider developer at  https://error.example.com/issues  ." in logger.messages['warning']

    def test_provider_unexpected_error_no_fallback(self, ie, logger, pot_cache, pot_request):
        director = PoTokenRequestDirector(logger=logger, cache=pot_cache)
        provider = UnexpectedErrorPTP(ie, logger, {})
        director.register_provider(provider)

        response = director.get_po_token(pot_request)
        assert response is None
        assert provider.request_called_times == 1
        assert "Unexpected error when fetching PO Token from \"unexpected_error\" provider: ValueError('an unexpected error occurred'); please report this issue to the provider developer at  https://unexpected.example.com/issues  ." in logger.messages['error']

    def test_provider_unexpected_error_fallback(self, ie, logger, pot_cache, pot_request, pot_provider):
        # Should fallback to the next provider if the first one raises an unexpected error
        director = PoTokenRequestDirector(logger=logger, cache=pot_cache)
        provider = UnexpectedErrorPTP(ie, logger, {})
        director.register_provider(provider)
        director.register_provider(pot_provider)

        response = director.get_po_token(pot_request)
        assert response == EXAMPLE_PO_TOKEN
        assert provider.request_called_times == 1
        assert pot_provider.request_called_times == 1
        assert "Unexpected error when fetching PO Token from \"unexpected_error\" provider: ValueError('an unexpected error occurred'); please report this issue to the provider developer at  https://unexpected.example.com/issues  ." in logger.messages['error']

    def test_invalid_po_token_response_type(self, ie, logger, pot_cache, pot_request, pot_provider):
        director = PoTokenRequestDirector(logger=logger, cache=pot_cache)
        provider = InvalidPTP(ie, logger, {})
        director.register_provider(provider)

        pot_request.video_id = 'invalid_type'

        response = director.get_po_token(pot_request)
        assert response is None
        assert provider.request_called_times == 1
        assert 'Invalid PO Token response received from "invalid" provider: invalid-response; please report this issue to the provider developer at  https://invalid.example.com/issues  .' in logger.messages['error']

        # Should fallback to next available provider
        director.register_provider(pot_provider)
        response = director.get_po_token(pot_request)
        assert response == EXAMPLE_PO_TOKEN
        assert provider.request_called_times == 2
        assert pot_provider.request_called_times == 1

    def test_invalid_po_token_response(self, ie, logger, pot_cache, pot_request, pot_provider):
        director = PoTokenRequestDirector(logger=logger, cache=pot_cache)
        provider = InvalidPTP(ie, logger, {})
        director.register_provider(provider)

        response = director.get_po_token(pot_request)
        assert response is None
        assert provider.request_called_times == 1
        assert "Invalid PO Token response received from \"invalid\" provider: PoTokenResponse(po_token='example-token?', expires_at='123'); please report this issue to the provider developer at  https://invalid.example.com/issues  ." in logger.messages['error']

        # Should fallback to next available provider
        director.register_provider(pot_provider)
        response = director.get_po_token(pot_request)
        assert response == EXAMPLE_PO_TOKEN
        assert provider.request_called_times == 2
        assert pot_provider.request_called_times == 1

    def test_copy_request_provider(self, ie, logger, pot_cache, pot_request):

        class BadProviderPTP(BaseMockPoTokenProvider):
            _SUPPORTED_CONTEXTS = None
            _SUPPORTED_CLIENTS = None

            def _real_request_pot(self, request: PoTokenRequest) -> PoTokenResponse:
                # Providers should not modify the request object, but we should guard against it
                request.video_id = 'bad'
                raise PoTokenProviderRejectedRequest('bad request')

        class GoodProviderPTP(BaseMockPoTokenProvider):
            _SUPPORTED_CONTEXTS = None
            _SUPPORTED_CLIENTS = None

            def _real_request_pot(self, request: PoTokenRequest) -> PoTokenResponse:
                return PoTokenResponse(base64.urlsafe_b64encode(request.video_id.encode()).decode())

        director = PoTokenRequestDirector(logger=logger, cache=pot_cache)

        bad_provider = BadProviderPTP(ie, logger, {})
        good_provider = GoodProviderPTP(ie, logger, {})

        director.register_provider(bad_provider)
        director.register_provider(good_provider)

        pot_request.video_id = 'good'
        response = director.get_po_token(pot_request)
        assert response == base64.urlsafe_b64encode(b'good').decode()
        assert bad_provider.request_called_times == 1
        assert good_provider.request_called_times == 1
        assert pot_request.video_id == 'good'


@pytest.mark.parametrize('spec, expected', [
    (None, False),
    (PoTokenCacheSpec(key_bindings={'v': 'video-id'}, default_ttl=60, write_policy=None), False),  # type: ignore
    (PoTokenCacheSpec(key_bindings={'v': 'video-id'}, default_ttl='invalid'), False),  # type: ignore
    (PoTokenCacheSpec(key_bindings='invalid', default_ttl=60), False),  # type: ignore
    (PoTokenCacheSpec(key_bindings={2: 'video-id'}, default_ttl=60), False),  # type: ignore
    (PoTokenCacheSpec(key_bindings={'v': 2}, default_ttl=60), False),  # type: ignore
    (PoTokenCacheSpec(key_bindings={'v': None}, default_ttl=60), False),  # type: ignore

    (PoTokenCacheSpec(key_bindings={'v': 'video_id', 'e': None}, default_ttl=60), True),
    (PoTokenCacheSpec(key_bindings={'v': 'video_id'}, default_ttl=60, write_policy=CacheProviderWritePolicy.WRITE_FIRST), True),
])
def test_validate_cache_spec(spec, expected):
    assert validate_cache_spec(spec) == expected


@pytest.mark.parametrize('po_token', [
    'invalid-token?',
    '123',
])
def test_clean_pot_fail(po_token):
    with pytest.raises(ValueError, match='Invalid PO Token'):
        clean_pot(po_token)


@pytest.mark.parametrize('po_token,expected', [
    ('TwAA/+8=', 'TwAA_-8='),
    ('TwAA%5F%2D9VA6Q92v%5FvEQ4==?extra-param=2', 'TwAA_-9VA6Q92v_vEQ4='),
])
def test_clean_pot(po_token, expected):
    assert clean_pot(po_token) == expected


@pytest.mark.parametrize(
    'response, expected',
    [
        (None, False),
        (PoTokenResponse(None), False),
        (PoTokenResponse(1), False),
        (PoTokenResponse('invalid-token?'), False),
        (PoTokenResponse(EXAMPLE_PO_TOKEN, expires_at='abc'), False),  # type: ignore
        (PoTokenResponse(EXAMPLE_PO_TOKEN, expires_at=100), False),
        (PoTokenResponse(EXAMPLE_PO_TOKEN, expires_at=time.time() + 10000.0), False),  # type: ignore
        (PoTokenResponse(EXAMPLE_PO_TOKEN), True),
        (PoTokenResponse(EXAMPLE_PO_TOKEN, expires_at=-1), True),
        (PoTokenResponse(EXAMPLE_PO_TOKEN, expires_at=0), True),
        (PoTokenResponse(EXAMPLE_PO_TOKEN, expires_at=int(time.time()) + 10000), True),
    ],
)
def test_validate_pot_response(response, expected):
    assert validate_response(response) == expected


def test_built_in_provider(ie, logger):
    class BuiltinProviderDefaultT(BuiltinIEContentProvider, suffix='T'):
        def is_available(self):
            return True

    class BuiltinProviderCustomNameT(BuiltinIEContentProvider, suffix='T'):
        PROVIDER_NAME = 'CustomName'

        def is_available(self):
            return True

    class ExternalProviderDefaultT(IEContentProvider, suffix='T'):
        def is_available(self):
            return True

    class ExternalProviderCustomT(IEContentProvider, suffix='T'):
        PROVIDER_NAME = 'custom'
        PROVIDER_VERSION = '5.4b2'

        def is_available(self):
            return True

    class ExternalProviderUnavailableT(IEContentProvider, suffix='T'):
        def is_available(self) -> bool:
            return False

    class BuiltinProviderUnavailableT(IEContentProvider, suffix='T'):
        def is_available(self) -> bool:
            return False

    built_in_default = BuiltinProviderDefaultT(ie=ie, logger=logger, settings={})
    built_in_custom_name = BuiltinProviderCustomNameT(ie=ie, logger=logger, settings={})
    built_in_unavailable = BuiltinProviderUnavailableT(ie=ie, logger=logger, settings={})
    external_default = ExternalProviderDefaultT(ie=ie, logger=logger, settings={})
    external_custom = ExternalProviderCustomT(ie=ie, logger=logger, settings={})
    external_unavailable = ExternalProviderUnavailableT(ie=ie, logger=logger, settings={})

    assert provider_display_list([]) == 'none'
    assert provider_display_list([built_in_default]) == 'BuiltinProviderDefault'
    assert provider_display_list([external_unavailable]) == 'ExternalProviderUnavailable-0.0.0 (external, unavailable)'
    assert provider_display_list([
        built_in_default,
        built_in_custom_name,
        external_default,
        external_custom,
        external_unavailable,
        built_in_unavailable],
    ) == 'BuiltinProviderDefault, CustomName, ExternalProviderDefault-0.0.0 (external), custom-5.4b2 (external), ExternalProviderUnavailable-0.0.0 (external, unavailable), BuiltinProviderUnavailable-0.0.0 (external, unavailable)'



from .common import InfoExtractor
from ..utils import (
    try_get,
    unified_strdate,
)


class PlanetMarathiIE(InfoExtractor):
    _VALID_URL = r'https?://(?:www\.)?planetmarathi\.com/titles/(?P<id>[^/#&?$]+)'
    _TESTS = [{
        'url': 'https://www.planetmarathi.com/titles/ek-unad-divas',
        'playlist_mincount': 2,
        'info_dict': {
            'id': 'ek-unad-divas',
        },
        'playlist': [{
            'info_dict': {
                'id': 'ASSETS-MOVIE-ASSET-01_ek-unad-divas',
                'ext': 'mp4',
                'title': 'ek unad divas',
                'alt_title': 'चित्रपट',
                'description': 'md5:41c7ed6b041c2fea9820a3f3125bd881',
                'episode_number': 1,
                'duration': 5539,
                'upload_date': '20210829',
            },
        }],  # Trailer skipped
    }, {
        'url': 'https://www.planetmarathi.com/titles/baap-beep-baap-season-1',
        'playlist_mincount': 10,
        'info_dict': {
            'id': 'baap-beep-baap-season-1',
        },
        'playlist': [{
            'info_dict': {
                'id': 'ASSETS-CHARACTER-PROFILE-SEASON-01-ASSET-01_baap-beep-baap-season-1',
                'ext': 'mp4',
                'title': 'Manohar Kanhere',
                'alt_title': 'मनोहर कान्हेरे',
                'description': 'md5:285ed45d5c0ab5522cac9a043354ebc6',
                'season_number': 1,
                'episode_number': 1,
                'duration': 29,
                'upload_date': '20210829',
            },
        }],  # Trailers, Episodes, other Character profiles skipped
    }]

    def _real_extract(self, url):
        playlist_id = self._match_id(url)
        entries = []
        json_data = self._download_json(
            f'https://www.planetmarathi.com/api/v1/titles/{playlist_id}/assets', playlist_id)['assets']
        for asset in json_data:
            asset_title = asset['mediaAssetName']['en']
            if asset_title == 'Movie':
                asset_title = playlist_id.replace('-', ' ')
            asset_id = f'{asset["sk"]}_{playlist_id}'.replace('#', '-')
            formats, subtitles = self._extract_m3u8_formats_and_subtitles(asset['mediaAssetURL'], asset_id)
            entries.append({
                'id': asset_id,
                'title': asset_title,
                'alt_title': try_get(asset, lambda x: x['mediaAssetName']['mr']),
                'description': try_get(asset, lambda x: x['mediaAssetDescription']['en']),
                'season_number': asset.get('mediaAssetSeason'),
                'episode_number': asset.get('mediaAssetIndexForAssetType'),
                'duration': asset.get('mediaAssetDurationInSeconds'),
                'upload_date': unified_strdate(asset.get('created')),
                'formats': formats,
                'subtitles': subtitles,
            })
        return self.playlist_result(entries, playlist_id=playlist_id)
